# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains code to build a ViP-DeepLab decoder.

Reference:
  - [ViP-DeepLab: Learning Visual Perception with Depth-aware Video
      Panoptic Segmentation](https://arxiv.org/abs/2012.05258)
"""
import tensorflow as tf

from deeplab2 import common
from deeplab2.model import utils
from deeplab2.model.decoder import panoptic_deeplab

layers = tf.keras.layers


class ViPDeepLabDecoder(layers.Layer):
    """A ViP-DeepLab decoder layer.

    This layer takes low- and high-level features as input and uses a dual-ASPP
    and dual-decoder structure to aggregate features for semantic and instance
    segmentation. On top of the decoders, three heads are used to predict semantic
    segmentation, instance center probabilities, and instance center regression
    per pixel. It also has a branch to predict the next-frame instance center
    regression. Different from the ViP-DeepLab paper which uses Cascade-ASPP, this
    reimplementation only uses ASPP.
    """

    def __init__(self,
                 decoder_options,
                 vip_deeplab_options,
                 bn_layer=tf.keras.layers.BatchNormalization):
        """Initializes a ViP-DeepLab decoder.

        Args:
          decoder_options: Decoder options as defined in config_pb2.DecoderOptions.
          vip_deeplab_options: Model options as defined in
            config_pb2.ModelOptions.ViPDeeplabOptions.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
        """
        super(ViPDeepLabDecoder, self).__init__(name='ViPDeepLab')

        low_level_feature_keys = [
            item.feature_key for item in vip_deeplab_options.low_level
        ]
        low_level_channels_project = [
            item.channels_project for item in vip_deeplab_options.low_level
        ]

        self._semantic_decoder = panoptic_deeplab.PanopticDeepLabSingleDecoder(
            high_level_feature_name=decoder_options.feature_key,
            low_level_feature_names=low_level_feature_keys,
            low_level_channels_project=low_level_channels_project,
            aspp_output_channels=decoder_options.aspp_channels,
            decoder_output_channels=decoder_options.decoder_channels,
            atrous_rates=decoder_options.atrous_rates,
            name='semantic_decoder',
            aspp_use_only_1x1_proj_conv=decoder_options.aspp_use_only_1x1_proj_conv,
            decoder_conv_type=decoder_options.decoder_conv_type,
            bn_layer=bn_layer)
        self._semantic_head = panoptic_deeplab.PanopticDeepLabSingleHead(
            vip_deeplab_options.semantic_head.head_channels,
            vip_deeplab_options.semantic_head.output_channels,
            common.PRED_SEMANTIC_LOGITS_KEY,
            name='semantic_head',
            conv_type=vip_deeplab_options.semantic_head.head_conv_type,
            bn_layer=bn_layer)

        self._depth_head = None
        if vip_deeplab_options.HasField('depth_head'):
            self._depth_head = panoptic_deeplab.PanopticDeepLabSingleHead(
                vip_deeplab_options.depth_head.head_channels,
                vip_deeplab_options.depth_head.output_channels,
                common.PRED_DEPTH_KEY,
                name='depth_head',
                conv_type=vip_deeplab_options.depth_head.head_conv_type,
                bn_layer=bn_layer)
            self._max_depth = (
                vip_deeplab_options.depth_head.max_value_after_activation)
            self._min_depth = (
                vip_deeplab_options.depth_head.min_value_after_activation)

        self._instance_decoder = None
        self._instance_center_head = None
        self._instance_regression_head = None
        self._next_instance_decoder = None
        self._next_instance_regression_head = None

        if vip_deeplab_options.instance.enable:
            if vip_deeplab_options.instance.low_level_override:
                low_level_options = vip_deeplab_options.instance.low_level_override
            else:
                low_level_options = vip_deeplab_options.low_level

            # If instance_decoder is set, use those options; otherwise reuse the
            # architecture as defined for the semantic decoder.
            if vip_deeplab_options.instance.HasField('instance_decoder_override'):
                decoder_options = (
                    vip_deeplab_options.instance.instance_decoder_override)

            low_level_feature_keys = [item.feature_key for item in low_level_options]
            low_level_channels_project = [
                item.channels_project for item in low_level_options
            ]

            self._instance_decoder = panoptic_deeplab.PanopticDeepLabSingleDecoder(
                high_level_feature_name=decoder_options.feature_key,
                low_level_feature_names=low_level_feature_keys,
                low_level_channels_project=low_level_channels_project,
                aspp_output_channels=decoder_options.aspp_channels,
                decoder_output_channels=decoder_options.decoder_channels,
                atrous_rates=decoder_options.atrous_rates,
                name='instance_decoder',
                aspp_use_only_1x1_proj_conv=(
                    decoder_options.aspp_use_only_1x1_proj_conv),
                decoder_conv_type=decoder_options.decoder_conv_type,
                bn_layer=bn_layer)
            self._instance_center_head = panoptic_deeplab.PanopticDeepLabSingleHead(
                vip_deeplab_options.instance.center_head.head_channels,
                vip_deeplab_options.instance.center_head.output_channels,
                common.PRED_CENTER_HEATMAP_KEY,
                name='instance_center_head',
                conv_type=(vip_deeplab_options.instance.center_head.head_conv_type),
                bn_layer=bn_layer)
            self._instance_regression_head = (
                panoptic_deeplab.PanopticDeepLabSingleHead(
                    vip_deeplab_options.instance.regression_head.head_channels,
                    vip_deeplab_options.instance.regression_head.output_channels,
                    common.PRED_OFFSET_MAP_KEY,
                    name='instance_regression_head',
                    conv_type=(
                        vip_deeplab_options.instance.regression_head.head_conv_type),
                    bn_layer=bn_layer))

            if vip_deeplab_options.instance.HasField('next_regression_head'):
                self._next_instance_decoder = (
                    panoptic_deeplab.PanopticDeepLabSingleDecoder(
                        high_level_feature_name=decoder_options.feature_key,
                        low_level_feature_names=low_level_feature_keys,
                        low_level_channels_project=low_level_channels_project,
                        aspp_output_channels=decoder_options.aspp_channels,
                        decoder_output_channels=decoder_options.decoder_channels,
                        atrous_rates=decoder_options.atrous_rates,
                        name='next_instance_decoder',
                        aspp_use_only_1x1_proj_conv=(
                            decoder_options.aspp_use_only_1x1_proj_conv),
                        decoder_conv_type=decoder_options.decoder_conv_type,
                        bn_layer=bn_layer))
                self._next_instance_regression_head = (
                    panoptic_deeplab.PanopticDeepLabSingleHead(
                        (vip_deeplab_options.instance.next_regression_head.head_channels
                         ), (vip_deeplab_options.instance.next_regression_head
                             .output_channels),
                        common.PRED_NEXT_OFFSET_MAP_KEY,
                        name='next_instance_regression_head',
                        conv_type=(vip_deeplab_options.instance.next_regression_head
                                   .head_conv_type),
                        bn_layer=bn_layer))
                self._next_high_level_feature_name = decoder_options.feature_key

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layers to global average pooling."""
        self._semantic_decoder.reset_pooling_layer()
        if self._instance_decoder is not None:
            self._instance_decoder.reset_pooling_layer()
        if self._next_instance_decoder is not None:
            self._next_instance_decoder.reset_pooling_layer()

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layers.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layers.
        """
        self._semantic_decoder.set_pool_size(pool_size)
        if self._instance_decoder is not None:
            self._instance_decoder.set_pool_size(pool_size)
        if self._next_instance_decoder is not None:
            self._next_instance_decoder.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._semantic_decoder.get_pool_size()

    @property
    def checkpoint_items(self):
        items = {
            common.CKPT_SEMANTIC_DECODER:
                self._semantic_decoder,
            common.CKPT_SEMANTIC_HEAD_WITHOUT_LAST_LAYER:
                self._semantic_head.conv_block,
            common.CKPT_SEMANTIC_LAST_LAYER:
                self._semantic_head.final_conv
        }
        if self._instance_decoder is not None:
            instance_items = {
                common.CKPT_INSTANCE_DECODER:
                    self._instance_decoder,
                common.CKPT_INSTANCE_CENTER_HEAD_WITHOUT_LAST_LAYER:
                    self._instance_center_head.conv_block,
                common.CKPT_INSTANCE_CENTER_HEAD_LAST_LAYER:
                    self._instance_center_head.final_conv,
                common.CKPT_INSTANCE_REGRESSION_HEAD_WITHOUT_LAST_LAYER:
                    self._instance_regression_head.conv_block,
                common.CKPT_INSTANCE_REGRESSION_HEAD_LAST_LAYER:
                    self._instance_regression_head.final_conv,
            }
            items.update(instance_items)
        if self._next_instance_decoder is not None:
            next_instance_items = {
                common.CKPT_NEXT_INSTANCE_DECODER:
                    self._next_instance_decoder,
                common.CKPT_NEXT_INSTANCE_REGRESSION_HEAD_WITHOUT_LAST_LAYER:
                    self._next_instance_regression_head.conv_block,
                common.CKPT_NEXT_INSTANCE_REGRESSION_HEAD_LAST_LAYER:
                    self._next_instance_regression_head.final_conv,
            }
            items.update(next_instance_items)
        if self._depth_head is not None:
            depth_items = {
                common.CKPT_DEPTH_HEAD_WITHOUT_LAST_LAYER:
                    self._depth_head.conv_block,
                common.CKPT_DEPTH_HEAD_LAST_LAYER:
                    self._depth_head.final_conv
            }
            items.update(depth_items)
        return items

    def call(self, features, next_features, training=False):
        """Performs a forward pass.

        Args:
          features: An input dict of tf.Tensor with shape [batch, height, width,
            channels]. Different keys should point to different features extracted
            by the encoder, e.g. low-level or high-level features.
          next_features: An input dict of tf.Tensor similar to features. The
            features are computed with the next frame as input.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the results of the semantic segmentation head and
            depending on the configuration also of the instance segmentation head.
        """

        semantic_features = self._semantic_decoder(features, training=training)
        results = self._semantic_head(semantic_features, training=training)

        if self._depth_head is not None:
            feature_size = semantic_features.get_shape().as_list()[1:3]
            scaled_feature_size = utils.scale_mutable_sequence(feature_size, 2)
            depth_features = utils.resize_align_corners(semantic_features,
                                                        scaled_feature_size)
            depth_prediction = self._depth_head(depth_features)
            for pred_key, pred_value in depth_prediction.items():
                pred_value = self._min_depth + tf.sigmoid(pred_value) * (
                        self._max_depth - self._min_depth)
                depth_prediction[pred_key] = pred_value
            results.update(depth_prediction)

        if self._instance_decoder is not None:
            instance_features = self._instance_decoder(features, training=training)
            instance_center_predictions = self._instance_center_head(
                instance_features, training=training)
            instance_regression_predictions = self._instance_regression_head(
                instance_features, training=training)

            if results.keys() & instance_center_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'center branch overlap. Please use unique keys.')
            results.update(instance_center_predictions)

            if results.keys() & instance_regression_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'regression branch overlap. Please use unique keys.')
            results.update(instance_regression_predictions)

        if self._next_instance_decoder is not None:
            # We update the high level features in next_features with the concated
            # features of the high level features in both features and next_features.
            high_level_feature_name = self._next_high_level_feature_name
            high_level_features = features[high_level_feature_name]
            next_high_level_features = next_features[high_level_feature_name]
            next_high_level_features = tf.concat(
                [high_level_features, next_high_level_features], axis=3)
            # Create a new dict for next_features to keep the original next_features
            # untouched for model exporting.
            new_next_features = dict()
            for key in next_features:
                if key == high_level_feature_name:
                    new_next_features[key] = next_high_level_features
                else:
                    new_next_features[key] = next_features[key]
            next_regression_features = self._next_instance_decoder(
                new_next_features, training=training)
            next_regression_predictions = self._next_instance_regression_head(
                next_regression_features, training=training)
            if results.keys() & next_regression_predictions.keys():
                raise ValueError('The keys of the next regresion branch overlap.'
                                 'Please use unique keys.')
            results.update(next_regression_predictions)

        return results



import sys
import torch
from math import sqrt
import torch.nn.functional as F
from lop.utils.AdamGnT import AdamGnT


class GnT(object):
    """
    Generate-and-Test algorithm for feed forward neural networks, based on
    maturity-threshold based replacement
    """
    def __init__(
            self,
            net,
            hidden_activation,
            opt,
            decay_rate=0.99,
            replacement_rate=1e-4,
            init='kaiming',
            device="cpu",
            maturity_threshold=20,
            util_type='contribution',
            loss_func=F.mse_loss,
            accumulate=False,
    ):
        super(GnT, self).__init__()
        self.device = device
        self.net = net
        self.num_hidden_layers = int(len(self.net) / 2)
        self.loss_func = loss_func
        self.accumulate = accumulate

        self.opt = opt
        self.opt_type = 'sgd'
        if isinstance(self.opt, AdamGnT):
            self.opt_type = 'adam'

        """
        Define the hyper-parameters of the algorithm
        """
        self.replacement_rate = replacement_rate
        self.decay_rate = decay_rate
        self.maturity_threshold = maturity_threshold
        self.util_type = util_type

        """
        Utility of all features/neurons
        """
        self.util = [torch.zeros(self.net[i * 2].out_features).to(self.device) for i in range(self.num_hidden_layers)]
        self.bias_corrected_util = \
            [torch.zeros(self.net[i * 2].out_features).to(self.device) for i in range(self.num_hidden_layers)]
        self.ages = [torch.zeros(self.net[i * 2].out_features).to(self.device) for i in range(self.num_hidden_layers)]
        self.m = torch.nn.Softmax(dim=1)
        self.mean_feature_act = [torch.zeros(self.net[i * 2].out_features).to(self.device) for i in range(self.num_hidden_layers)]
        self.accumulated_num_features_to_replace = [0 for i in range(self.num_hidden_layers)]

        """
        Calculate uniform distribution's bound for random feature initialization
        """
        if hidden_activation == 'selu':
            init = 'lecun'
        self.bounds = self.compute_bounds(hidden_activation=hidden_activation, init=init)

    def compute_bounds(self, hidden_activation, init='kaiming'):
        if hidden_activation in ['swish', 'elu']:
            hidden_activation = 'relu'
        if init == 'default':
            bounds = [sqrt(1 / self.net[i * 2].in_features) for i in range(self.num_hidden_layers)]
        elif init == 'xavier':
            bounds = [torch.nn.init.calculate_gain(nonlinearity=hidden_activation) *
                      sqrt(6 / (self.net[i * 2].in_features + self.net[i * 2].out_features)) for i in
                      range(self.num_hidden_layers)]
        elif init == 'lecun':
            bounds = [sqrt(3 / self.net[i * 2].in_features) for i in range(self.num_hidden_layers)]
        else:
            bounds = [torch.nn.init.calculate_gain(nonlinearity=hidden_activation) *
                      sqrt(3 / self.net[i * 2].in_features) for i in range(self.num_hidden_layers)]
        bounds.append(1 * sqrt(3 / self.net[self.num_hidden_layers * 2].in_features))
        return bounds

    def update_utility(self, layer_idx=0, features=None, next_features=None):
        with torch.no_grad():
            self.util[layer_idx] *= self.decay_rate
            """
            Adam-style bias correction
            """
            bias_correction = 1 - self.decay_rate ** self.ages[layer_idx]

            self.mean_feature_act[layer_idx] *= self.decay_rate
            self.mean_feature_act[layer_idx] -= - (1 - self.decay_rate) * features.mean(dim=0)
            bias_corrected_act = self.mean_feature_act[layer_idx] / bias_correction

            current_layer = self.net[layer_idx * 2]
            next_layer = self.net[layer_idx * 2 + 2]
            output_wight_mag = next_layer.weight.data.abs().mean(dim=0)
            input_wight_mag = current_layer.weight.data.abs().mean(dim=1)

            if self.util_type == 'weight':
                new_util = output_wight_mag
            elif self.util_type == 'contribution':
                new_util = output_wight_mag * features.abs().mean(dim=0)
            elif self.util_type == 'adaptation':
                new_util = 1 / input_wight_mag
            elif self.util_type == 'zero_contribution':
                new_util = output_wight_mag * (features - bias_corrected_act).abs().mean(dim=0)
            elif self.util_type == 'adaptable_contribution':
                new_util = output_wight_mag * (features - bias_corrected_act).abs().mean(dim=0) / input_wight_mag
            elif self.util_type == 'feature_by_input':
                input_wight_mag = self.net[layer_idx * 2].weight.data.abs().mean(dim=1)
                new_util = (features - bias_corrected_act).abs().mean(dim=0) / input_wight_mag
            else:
                new_util = 0

            self.util[layer_idx] += (1 - self.decay_rate) * new_util

            """
            Adam-style bias correction
            """
            self.bias_corrected_util[layer_idx] = self.util[layer_idx] / bias_correction

            if self.util_type == 'random':
                self.bias_corrected_util[layer_idx] = torch.rand(self.util[layer_idx].shape)

    def test_features(self, features):
        """
        Args:
            features: Activation values in the neural network
        Returns:
            Features to replace in each layer, Number of features to replace in each layer
        """
        features_to_replace = [torch.empty(0, dtype=torch.long).to(self.device) for _ in range(self.num_hidden_layers)]
        num_features_to_replace = [0 for _ in range(self.num_hidden_layers)]
        if self.replacement_rate == 0:
            return features_to_replace, num_features_to_replace
        for i in range(self.num_hidden_layers):
            self.ages[i] += 1
            """
            Update feature utility
            """
            self.update_utility(layer_idx=i, features=features[i])
            """
            Find the no. of features to replace
            """
            eligible_feature_indices = torch.where(self.ages[i] > self.maturity_threshold)[0]
            if eligible_feature_indices.shape[0] == 0:
                continue
            num_new_features_to_replace = self.replacement_rate * eligible_feature_indices.shape[0]
            self.accumulated_num_features_to_replace[i] += num_new_features_to_replace

            """
            Case when the number of features to be replaced is between 0 and 1.
            """
            if self.accumulate:
                num_new_features_to_replace = int(self.accumulated_num_features_to_replace[i])
                self.accumulated_num_features_to_replace[i] -= num_new_features_to_replace
            else:
                if num_new_features_to_replace < 1:
                    if torch.rand(1) <= num_new_features_to_replace:
                        num_new_features_to_replace = 1
                num_new_features_to_replace = int(num_new_features_to_replace)

            if num_new_features_to_replace == 0:
                continue

            """
            Find features to replace in the current layer
            """
            new_features_to_replace = torch.topk(-self.bias_corrected_util[i][eligible_feature_indices],
                                                 num_new_features_to_replace)[1]
            new_features_to_replace = eligible_feature_indices[new_features_to_replace]

            """
            Initialize utility for new features
            """
            self.util[i][new_features_to_replace] = 0
            self.mean_feature_act[i][new_features_to_replace] = 0.

            features_to_replace[i] = new_features_to_replace
            num_features_to_replace[i] = num_new_features_to_replace

        return features_to_replace, num_features_to_replace

    def gen_new_features(self, features_to_replace, num_features_to_replace):
        """
        Generate new features: Reset input and output weights for low utility features
        """
        with torch.no_grad():
            for i in range(self.num_hidden_layers):
                if num_features_to_replace[i] == 0:
                    continue
                current_layer = self.net[i * 2]
                next_layer = self.net[i * 2 + 2]
                current_layer.weight.data[features_to_replace[i], :] *= 0.0
                # noinspection PyArgumentList
                current_layer.weight.data[features_to_replace[i], :] += \
                    torch.empty(num_features_to_replace[i], current_layer.in_features).uniform_(
                        -self.bounds[i], self.bounds[i]).to(self.device)
                current_layer.bias.data[features_to_replace[i]] *= 0
                """
                # Update bias to correct for the removed features and set the outgoing weights and ages to zero
                """
                next_layer.bias.data += (next_layer.weight.data[:, features_to_replace[i]] * \
                                                self.mean_feature_act[i][features_to_replace[i]] / \
                                                (1 - self.decay_rate ** self.ages[i][features_to_replace[i]])).sum(dim=1)
                next_layer.weight.data[:, features_to_replace[i]] = 0
                self.ages[i][features_to_replace[i]] = 0

    def update_optim_params(self, features_to_replace, num_features_to_replace):
        """
        Update Optimizer's state
        """
        if self.opt_type == 'adam':
            for i in range(self.num_hidden_layers):
                # input weights
                if num_features_to_replace == 0:
                    continue
                self.opt.state[self.net[i * 2].weight]['exp_avg'][features_to_replace[i], :] = 0.0
                self.opt.state[self.net[i * 2].bias]['exp_avg'][features_to_replace[i]] = 0.0
                self.opt.state[self.net[i * 2].weight]['exp_avg_sq'][features_to_replace[i], :] = 0.0
                self.opt.state[self.net[i * 2].bias]['exp_avg_sq'][features_to_replace[i]] = 0.0
                self.opt.state[self.net[i * 2].weight]['step'][features_to_replace[i], :] = 0
                self.opt.state[self.net[i * 2].bias]['step'][features_to_replace[i]] = 0
                # output weights
                self.opt.state[self.net[i * 2 + 2].weight]['exp_avg'][:, features_to_replace[i]] = 0.0
                self.opt.state[self.net[i * 2 + 2].weight]['exp_avg_sq'][:, features_to_replace[i]] = 0.0
                self.opt.state[self.net[i * 2 + 2].weight]['step'][:, features_to_replace[i]] = 0

    def gen_and_test(self, features):
        """
        Perform generate-and-test
        :param features: activation of hidden units in the neural network
        """
        if not isinstance(features, list):
            print('features passed to generate-and-test should be a list')
            sys.exit()
        features_to_replace, num_features_to_replace = self.test_features(features=features)
        self.gen_new_features(features_to_replace, num_features_to_replace)
        self.update_optim_params(features_to_replace, num_features_to_replace)



"""
Templates for script that will be written to disk for use by a user or another
program.
"""
import ubelt as ub


def train_time_helper_scripts(dpath, train_coco_fpath, vali_coco_fpath):
    """
    Example:
        >>> from geowatch.tasks.fusion.helper_scripts import *
        >>> dpath = ub.Path('.')
        >>> train_coco_fpath = ub.Path('train.kwcoco.zip')
        >>> vali_coco_fpath = ub.Path('vali.kwcoco.zip')
        >>> scripts = train_time_helper_scripts(dpath, train_coco_fpath, vali_coco_fpath)
        >>> print(f'scripts = {ub.urepr(scripts, nl=3, sv=1)}')
    """
    scripts = {}

    key = 'start_tensorboard'
    script = scripts[key] = {}
    script['fpath'] = dpath / f'{key}.sh'
    script['text'] = ub.codeblock(
        f'''
        #!/usr/bin/env bash
        tensorboard --logdir {dpath}
        ''')

    try:
        generating_fpath = __file__
    except NameError:
        generating_fpath = 'geowatch/tasks/fusion/helper_scripts.py'
    autogen_comment = f'# Generated by {generating_fpath}'

    key = 'draw_tensorboard'
    script = scripts[key] = {}
    script['fpath'] = dpath / f'{key}.sh'
    script['text'] = (ub.codeblock(
        fr'''
        #!/usr/bin/env bash
        {autogen_comment}

        # Update the main plots and stack them into a nice figure
        GEOWATCH_PREIMPORT=0 python -m geowatch.utils.lightning_ext.callbacks.tensorboard_plotter \
            {dpath}
        '''
    ))

    checkpoint_header_part = ub.codeblock(
        fr'''
        #!/usr/bin/env bash
        {autogen_comment}

        # Device defaults to CPU, but the user can pass a GPU in
        # as the first argument.
        DEVICE=${{1:-"cpu"}}

        TRAIN_DPATH="{dpath}"
        echo $TRAIN_DPATH

        ### --- Choose Checkpoint --- ###

        # Find a checkpoint to evaluate
        # TODO: should add a geowatch helper for this
        CHECKPOINT_FPATH=$(python -c "if 1:
            import pathlib
            train_dpath = pathlib.Path('$TRAIN_DPATH')
            found = sorted((train_dpath / 'checkpoints').glob('*.ckpt'))
            found = [f for f in found if 'last.ckpt' not in str(f)]
            print(found[-1])
            ")
        echo "$CHECKPOINT_FPATH"

        ### --- Repackage Checkpoint --- ###

        # Convert it into a package, then get the name of that
        geowatch repackage "$CHECKPOINT_FPATH"

        PACKAGE_FPATH=$(python -c "if 1:
            import pathlib
            p = pathlib.Path('$CHECKPOINT_FPATH')
            found = list(p.parent.glob(p.stem + '*.pt'))
            print(found[-1])
        ")
        echo "$PACKAGE_FPATH"

        PACKAGE_NAME=$(python -c "if 1:
            import pathlib
            p = pathlib.Path('$PACKAGE_FPATH')
            print(p.stem.replace('.ckpt', ''))
        ")
        echo "$PACKAGE_NAME"
        ''')

    key = 'draw_train_batches'
    script = scripts[key] = {}
    script['fpath'] = dpath / f'{key}.sh'
    script['text'] = chr(10).join([
        checkpoint_header_part,
        ub.codeblock(
            fr'''
            ### --- Train Batch Prediction --- ###

            # Predict on the validation set
            export IGNORE_OFF_BY_ONE_STITCHING=1  # hack
            python -m geowatch.tasks.fusion.predict \
                --package_fpath "$PACKAGE_FPATH" \
                --test_dataset "{train_coco_fpath}" \
                --pred_dataset "$TRAIN_DPATH/monitor/train/preds/$PACKAGE_NAME/pred-$PACKAGE_NAME.kwcoco.zip" \
                --window_overlap 0 \
                --clear_annots=False \
                --test_with_annot_info=True \
                --use_centered_positives=True \
                --use_grid_positives=False \
                --use_grid_negatives=False \
                --draw_batches=True \
                --devices "$DEVICE"
            ''')
    ])

    key = 'draw_vali_batches'
    script = scripts[key] = {}
    script['fpath'] = dpath / f'{key}.sh'
    script['text'] = chr(10).join([
        checkpoint_header_part,
        ub.codeblock(
            fr'''
            ### --- Validation Batch Prediction --- ###

            # Predict on the validation set
            export IGNORE_OFF_BY_ONE_STITCHING=1  # hack
            python -m geowatch.tasks.fusion.predict \
                --package_fpath "$PACKAGE_FPATH" \
                --test_dataset "{vali_coco_fpath}" \
                --pred_dataset "$TRAIN_DPATH/monitor/vali/preds/$PACKAGE_NAME/pred-$PACKAGE_NAME.kwcoco.zip" \
                --window_overlap 0 \
                --clear_annots=False \
                --test_with_annot_info=True \
                --use_centered_positives=True \
                --use_grid_positives=False \
                --use_grid_negatives=False \
                --draw_batches=True \
                --devices "$DEVICE"
            ''')
    ])

    key = 'draw_train_dataset'
    script = scripts[key] = {}
    script['fpath'] = dpath / f'{key}.sh'
    script['text'] = chr(10).join([
        checkpoint_header_part,
        ub.codeblock(
            fr'''
            ### --- Train Full-Image Prediction (best run on a GPU) --- ###

            # Predict on the training set
            export IGNORE_OFF_BY_ONE_STITCHING=1  # hack
            python -m geowatch.tasks.fusion.predict \
                --package_fpath "$PACKAGE_FPATH" \
                --window_overlap 0 \
                --test_dataset "{train_coco_fpath}" \
                --pred_dataset "$TRAIN_DPATH/monitor/train/preds/$PACKAGE_NAME/pred-$PACKAGE_NAME.kwcoco.zip" \
                --clear_annots False \
                --devices "$DEVICE"

            # Visualize train predictions
            geowatch visualize "$TRAIN_DPATH/monitor/train/preds/$PACKAGE_NAME/pred-$PACKAGE_NAME.kwcoco.zip" --smart
            ''')
    ])

    key = 'draw_vali_dataset'
    script = scripts[key] = {}
    script['fpath'] = dpath / f'{key}.sh'
    script['text'] = chr(10).join([
        checkpoint_header_part,
        ub.codeblock(
            fr'''
            ### --- Validation Full-Image Prediction (best run on a GPU) --- ###

            # Predict on the validation set
            export IGNORE_OFF_BY_ONE_STITCHING=1  # hack
            python -m geowatch.tasks.fusion.predict \
                --package_fpath "$PACKAGE_FPATH" \
                --test_dataset "{vali_coco_fpath}" \
                --pred_dataset "$TRAIN_DPATH/monitor/vali/preds/$PACKAGE_NAME/pred-$PACKAGE_NAME.kwcoco.zip" \
                --window_overlap 0 \
                --clear_annots=False \
                --devices "$DEVICE"

            # Visualize vali predictions
            geowatch visualize $TRAIN_DPATH/monitor/vali/preds/$PACKAGE_NAME/pred-$PACKAGE_NAME.kwcoco.zip --smart
            ''')
    ])
    return scripts



import torch


class SampleUniform(object):

    def __init__(self, low, high):
        self.m = torch.distributions.uniform.Uniform(low, high)

    def __call__(self):
        return self.m.sample()



from PIL import Image, ImageDraw
import numpy as np
import shapely.geometry
import shapely.affinity

from . import functional


class AngleFieldInit(object):
    def __init__(self, line_width=1):
        self.line_width = line_width

    def __call__(self, image, polygons):
        size = (image.shape[0], image.shape[1])
        return init_angle_field(polygons, size, self.line_width)


def init_angle_field(polygons, shape, line_width=1):
    """
    Angle field {\theta_1} the tangent vector's angle for every pixel, specified on the polygon edges.
    Angle between 0 and pi.
    This is not invariant to symmetries.

    :param polygons:
    :param shape:
    :return: (angles: np.array((num_edge_pixels, ), dtype=np.uint8),
              mask: np.array((num_edge_pixels, 2), dtype=np.int))
    """
    assert isinstance(polygons, list), "polygons should be a list"
    if len(polygons):
        assert isinstance(polygons[0], shapely.geometry.Polygon), "polygon should be a shapely.geometry.Polygon"

    im = Image.new("L", (shape[1], shape[0]))
    im_px_access = im.load()  # NOQA
    draw = ImageDraw.Draw(im)

    for polygon in polygons:
        draw_linear_ring(draw, polygon.exterior, line_width)
        for interior in polygon.interiors:
            draw_linear_ring(draw, interior, line_width)

    # Convert image to numpy array
    array = np.array(im)
    return array


def draw_linear_ring(draw, linear_ring, line_width):
    # --- edges:
    coords = np.array(linear_ring)
    edge_vect_array = np.diff(coords, axis=0)
    edge_angle_array = np.angle(edge_vect_array[:, 1] + 1j * edge_vect_array[:, 0])  # ij coord sys
    neg_indices = np.where(edge_angle_array < 0)
    edge_angle_array[neg_indices] += np.pi

    first_uint8_angle = None
    for i in range(coords.shape[0] - 1):
        edge = (coords[i], coords[i + 1])
        angle = edge_angle_array[i]
        uint8_angle = int((255 * angle / np.pi).round())
        if first_uint8_angle is None:
            first_uint8_angle = uint8_angle
        line = [(edge[0][0], edge[0][1]), (edge[1][0], edge[1][1])]
        draw.line(line, fill=uint8_angle, width=line_width)
        functional.draw_circle(draw, line[0], radius=line_width / 2, fill=uint8_angle)

    # Add first vertex back on top (equals to last vertex too):
    functional.draw_circle(draw, line[1], radius=line_width / 2, fill=first_uint8_angle)



"""
DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc
"""
# xdoctest: +REQUIRES(env:DVC_DPATH)
# Run the following tests on real watch data if DVC is available
import os
from os.path import join
import kwcoco
import ubelt as ub
import kwimage
import numpy as np
from typing import List, Dict


def demo():
    # display with matplotlib
    DRAW = 1
    if DRAW:
        import kwplot
        kwplot.autompl()

    _default = ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc')
    dvc_dpath = os.environ.get('DVC_DPATH', _default)
    coco_fpath = join(dvc_dpath, 'drop1-S2-L8-aligned/data.kwcoco.json')
    dset = kwcoco.CocoDataset(coco_fpath)

    # Given an image
    img1_id = ub.peek(dset.imgs.keys())

    # Insetad of doing this
    annot_ids = dset.index.gid_to_aids[img1_id]
    img1 = dset.delayed_load(img1_id, channels='red|green|blue', space='video').finalize()
    dets1 = kwimage.Detections.from_coco_annots(
                    dset.annots(annot_ids).objs, dset=dset)

    if DRAW:
        canvas1 = kwimage.normalize_intensity(img1)
        canvas2 = dets1.draw_on(canvas1.copy())
        canvas3 = dets1.draw_on(np.zeros_like(canvas1))
        kwplot.imshow(canvas1, fnum=1, pnum=(3, 3, 1), title='kwcoco, orig, img only')
        kwplot.imshow(canvas2, fnum=1, pnum=(3, 3, 2), title='kwcoco, orig, img + annot')
        kwplot.imshow(canvas3, fnum=1, pnum=(3, 3, 3), title='kwcoco, orig, annot only')

    # This is the fix
    warp_vid_from_img = kwimage.Affine.coerce(dset.index.imgs[img1_id]['warp_img_to_vid'])
    dets1_fixed = dets1.warp(warp_vid_from_img)

    if DRAW:
        canvas1 = kwimage.normalize_intensity(img1)
        canvas2 = dets1_fixed.draw_on(canvas1.copy())
        canvas3 = dets1_fixed.draw_on(np.zeros_like(canvas1))
        kwplot.imshow(canvas1, fnum=1, pnum=(3, 3, 4), title='kwcoco, fix, img only')
        kwplot.imshow(canvas2, fnum=1, pnum=(3, 3, 5), title='kwcoco, fix, img + annot')
        kwplot.imshow(canvas3, fnum=1, pnum=(3, 3, 6), title='kwcoco, fix, annot only')

    # You may just want to do this
    import ndsampler
    sampler = ndsampler.CocoSampler(dset)

    # The sampler can load a sample
    sample       : Dict[str, object] = sampler.load_sample({'gids': [img1_id], 'channels': 'red|green|blue'})

    # Here are useful items in the sample, it contains more metadata than this
    frame_imdata : np.ndarray               = sample['im']
    frame_dets   : List[kwimage.Detections] = sample['annots']['frame_dets']

    # We loaded a sequence of images, so just take the first
    dets1 = frame_dets[0]
    img1 = frame_imdata[0]
    if DRAW:
        # display with matplotlib
        import kwplot
        kwplot.autompl()
        canvas1 = kwimage.normalize_intensity(img1)
        canvas2 = dets1.draw_on(canvas1.copy(), labels=False)
        canvas3 = dets1.draw_on(np.zeros_like(canvas1), labels=False)
        kwplot.imshow(canvas1, fnum=1, pnum=(3, 3, 7), title='ndsampler, img only')
        kwplot.imshow(canvas2, fnum=1, pnum=(3, 3, 8), title='ndsampler, img + annot')
        kwplot.imshow(canvas3, fnum=1, pnum=(3, 3, 9), title='ndsampler, annot only')


def demo2():
    # display with matplotlib
    DRAW = 1
    if DRAW:
        import kwplot
        kwplot.autompl()

    _default = ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc')
    dvc_dpath = os.environ.get('DVC_DPATH', _default)
    coco_fpath = join(dvc_dpath, 'drop1-S2-L8-aligned/data.kwcoco.json')
    dset = kwcoco.CocoDataset(coco_fpath)

    # Given an image
    video_id = ub.peek(dset.index.videos.keys())

    video_info = dset.index.videos[video_id]
    video_gids = dset.index.vidid_to_gids[video_id]

    vid_w = video_info['width']
    vid_h = video_info['height']

    rng = np.random.RandomState(0)
    chosen_gids = sorted(rng.choice(video_gids, size=3, replace=False))

    target = {
        # Right half of the video
        'space_slice': (slice(0, vid_h), slice(vid_w // 2, vid_w)),
        'gids': chosen_gids,
        'channels': 'red|green|blue'
    }

    import ndsampler
    sampler = ndsampler.CocoSampler(dset)

    # The sampler can load a sample
    sample       : Dict[str, object] = sampler.load_sample(target)

    # Here are useful items in the sample, it contains more metadata than this
    frame_imdata : np.ndarray               = sample['im']
    frame_dets   : List[kwimage.Detections] = sample['annots']['frame_dets']

    # We loaded a sequence of images, so just take the first

    if DRAW:
        vizlist = []
        for frame_idx in range(len(chosen_gids)):
            imdata = frame_imdata[frame_idx]
            dets = frame_dets[frame_idx]

            canvas = kwimage.normalize_intensity(imdata[:, :, 0:3])

            try:
                canvas = dets.draw_on(canvas, color='classes')
            except Exception:
                # Seems bugged, disable labels
                canvas = dets.draw_on(canvas, labels=False)

            vizlist.append(canvas)

        final_canvas = kwimage.stack_images(vizlist, axis=1, overlap=-10)  # negative overlap is a weird parameter for a pad I know. I don't know what I was thinking.
        kwplot.imshow(final_canvas)


def demo3():
    import torch
    # display with matplotlib
    DRAW = 1
    if DRAW:
        import kwplot
        kwplot.autompl()

    _default = ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc')
    dvc_dpath = os.environ.get('DVC_DPATH', _default)
    coco_fpath = join(dvc_dpath, 'drop1-S2-L8-aligned/combo_data.kwcoco.json')
    dset = kwcoco.CocoDataset(coco_fpath)

    # Given a video
    video_id = ub.peek(dset.index.videos.keys())
    video_gids = dset.index.vidid_to_gids[video_id]

    # Choose an image, lets build a pandas dataframe to make an informed choice

    df_rows = []
    for gid in video_gids:
        _imginfo = dset.index.imgs[gid]
        df_rows.append({
            'gid': gid,
            'sensor': _imginfo['sensor_coarse'],
            'num_annots': len(dset.index.gid_to_aids[gid])
        })
    import pandas as pd
    frames_df = pd.DataFrame(df_rows)
    candidates = frames_df[frames_df.sensor == 'S2'].sort_values('num_annots')
    image_id = candidates.gid.iloc[13]

    # Gether the annotations inside that image
    annot_ids = dset.index.gid_to_aids[image_id]

    # Create information about video space and image space
    # (Reminder: annotations are stored in image space)
    img_info = dset.index.imgs[image_id]
    img_dims = (img_info['height'], img_info['width'])

    video_info = dset.index.videos[video_id]
    vid_dims = (video_info['height'], video_info['width'])

    warp_vid_from_img = kwimage.Affine.coerce(dset.index.imgs[image_id]['warp_img_to_vid'])

    # Construct variants of the image and annotations
    imdata_in_imgspace = dset.delayed_load(image_id, channels='red|green|blue', space='image').finalize()
    imdata_in_vidspace = dset.delayed_load(image_id, channels='red|green|blue', space='video').finalize()

    dets_in_imgspace = kwimage.Detections.from_coco_annots(
        dset.annots(annot_ids).objs, dset=dset)
    dets_in_vidspace = dets_in_imgspace.warp(warp_vid_from_img)

    # Mask construction in image space
    segmentations = dets_in_imgspace.data['segmentations'].data
    category_ids = [dets_in_imgspace.classes.idx_to_id[cidx]
                    for cidx in dets_in_imgspace.data['class_idxs']]
    combined = []
    for sseg, cid in zip(segmentations, category_ids):
        assert cid > 0
        np_mask = sseg.to_mask(dims=img_dims).data.astype(float) * cid
        mask1 = torch.from_numpy(np_mask)
        combined.append(mask1.unsqueeze(0))
    overall_mask = torch.max(torch.cat(combined, dim=0), dim=0)[0]
    mask_in_imgspace = ((overall_mask > 1) * 255).numpy().astype(np.uint8)

    # Mask construction in video space
    segmentations = dets_in_vidspace.data['segmentations'].data
    category_ids = [dets_in_vidspace.classes.idx_to_id[cidx]
                    for cidx in dets_in_vidspace.data['class_idxs']]
    combined = []
    for sseg, cid in zip(segmentations, category_ids):
        assert cid > 0
        np_mask = sseg.to_mask(dims=vid_dims).data.astype(float) * cid
        mask1 = torch.from_numpy(np_mask)
        combined.append(mask1.unsqueeze(0))
    overall_mask = torch.max(torch.cat(combined, dim=0), dim=0)[0]
    mask_in_vidspace = ((overall_mask > 1) * 255).numpy().astype(np.uint8)

    if DRAW:
        vidspace_canvas_raw = kwimage.normalize_intensity(imdata_in_vidspace)
        imgspace_canvas_raw = kwimage.normalize_intensity(imdata_in_imgspace)
        print('imgspace_canvas_raw.shape = {!r}'.format(imgspace_canvas_raw.shape))
        print('vidspace_canvas_raw.shape = {!r}'.format(vidspace_canvas_raw.shape))

        vidspace_canvas = dets_in_imgspace.draw_on(vidspace_canvas_raw.copy(), color='orange', labels=['img-space'] * len(annot_ids))
        vidspace_canvas = dets_in_vidspace.draw_on(vidspace_canvas, color='dodgerblue', labels=['vid-space'] * len(annot_ids))

        imgspace_canvas = dets_in_imgspace.draw_on(imgspace_canvas_raw.copy(), color='orange', labels=['img-space'] * len(annot_ids))
        imgspace_canvas = dets_in_vidspace.draw_on(imgspace_canvas, color='dodgerblue', labels=['vid-space'] * len(annot_ids))

        heatmask_in_vidspace = kwimage.ensure_alpha_channel(mask_in_vidspace, alpha=0.4)
        heatmask_in_imgspace = kwimage.ensure_alpha_channel(mask_in_imgspace, alpha=0.4)
        overlay_in_vidspace = kwimage.overlay_alpha_layers([heatmask_in_vidspace, vidspace_canvas_raw])
        overlay_in_imgspace = kwimage.overlay_alpha_layers([heatmask_in_imgspace, imgspace_canvas_raw])

        pnum_ = kwplot.PlotNums(nRows=3, nCols=2)
        kwplot.imshow(imgspace_canvas, fnum=1, pnum=pnum_(), title='img-space-canvas', show_ticks=1)
        kwplot.imshow(vidspace_canvas, fnum=1, pnum=pnum_(), title='vid-space-canvas', show_ticks=1)

        kwplot.imshow(mask_in_imgspace, fnum=1, pnum=pnum_(), title='img-space-mask', show_ticks=1)
        kwplot.imshow(mask_in_vidspace, fnum=1, pnum=pnum_(), title='vid-space-mask', show_ticks=1)

        kwplot.imshow(overlay_in_imgspace, fnum=1, pnum=pnum_(), title='img-space-overlay', show_ticks=1)
        kwplot.imshow(overlay_in_vidspace, fnum=1, pnum=pnum_(), title='vid-space-overlay', show_ticks=1)


def demo4():
    import torch
    # display with matplotlib
    DRAW = 1
    if DRAW:
        import kwplot
        kwplot.autompl()

    _default = ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc')
    dvc_dpath = os.environ.get('DVC_DPATH', _default)
    coco_fpath = join(dvc_dpath, 'drop1-S2-L8-aligned/combo_data.kwcoco.json')
    dset = kwcoco.CocoDataset(coco_fpath)

    # Given a video
    video_id = ub.peek(dset.index.videos.keys())
    video_gids = dset.index.vidid_to_gids[video_id]

    # Choose an image, lets build a pandas dataframe to make an informed choice

    df_rows = []
    for gid in video_gids:
        _imginfo = dset.index.imgs[gid]
        df_rows.append({
            'gid': gid,
            'sensor': _imginfo['sensor_coarse'],
            'num_annots': len(dset.index.gid_to_aids[gid])
        })
    import pandas as pd
    frames_df = pd.DataFrame(df_rows)
    candidates = frames_df[frames_df.sensor == 'S2'].sort_values('num_annots')
    image_id = candidates.gid.iloc[13]

    # Gether the annotations inside that image
    annot_ids = dset.index.gid_to_aids[image_id]

    # Create information about video space and image space
    # (Reminder: annotations are stored in image space)
    img_info = dset.index.imgs[image_id]
    img_dims = (img_info['height'], img_info['width'])

    video_info = dset.index.videos[video_id]
    vid_dims = (video_info['height'], video_info['width'])

    warp_vid_from_img = kwimage.Affine.coerce(dset.index.imgs[image_id]['warp_img_to_vid'])

    # Construct variants of the image and annotations
    imdata_in_imgspace = dset.delayed_load(image_id, channels='red|green|blue', space='image').finalize()
    imdata_in_vidspace = dset.delayed_load(image_id, channels='red|green|blue', space='video').finalize()

    dets_in_imgspace = kwimage.Detections.from_coco_annots(
        dset.annots(annot_ids).objs, dset=dset)
    dets_in_vidspace = dets_in_imgspace.warp(warp_vid_from_img)

    # Mask construction in image space
    segmentations = dets_in_imgspace.data['segmentations'].data
    category_ids = [dets_in_imgspace.classes.idx_to_id[cidx]
                    for cidx in dets_in_imgspace.data['class_idxs']]
    combined = []
    for sseg, cid in zip(segmentations, category_ids):
        assert cid > 0
        np_mask = sseg.to_mask(dims=img_dims).data.astype(float) * cid
        mask1 = torch.from_numpy(np_mask)
        combined.append(mask1.unsqueeze(0))
    overall_mask = torch.max(torch.cat(combined, dim=0), dim=0)[0]
    mask_in_imgspace = ((overall_mask > 1) * 255).numpy().astype(np.uint8)

    # Mask construction in video space
    segmentations = dets_in_vidspace.data['segmentations'].data
    category_ids = [dets_in_vidspace.classes.idx_to_id[cidx]
                    for cidx in dets_in_vidspace.data['class_idxs']]
    combined = []
    for sseg, cid in zip(segmentations, category_ids):
        assert cid > 0
        np_mask = sseg.to_mask(dims=vid_dims).data.astype(float) * cid
        mask1 = torch.from_numpy(np_mask)
        combined.append(mask1.unsqueeze(0))
    overall_mask = torch.max(torch.cat(combined, dim=0), dim=0)[0]
    mask_in_vidspace = ((overall_mask > 1) * 255).numpy().astype(np.uint8)

    if DRAW:
        vidspace_canvas_raw = kwimage.normalize_intensity(imdata_in_vidspace)
        imgspace_canvas_raw = kwimage.normalize_intensity(imdata_in_imgspace)
        print('imgspace_canvas_raw.shape = {!r}'.format(imgspace_canvas_raw.shape))
        print('vidspace_canvas_raw.shape = {!r}'.format(vidspace_canvas_raw.shape))

        vidspace_canvas = dets_in_imgspace.draw_on(vidspace_canvas_raw.copy(), color='orange', labels=['img-space'] * len(annot_ids))
        vidspace_canvas = dets_in_vidspace.draw_on(vidspace_canvas, color='dodgerblue', labels=['vid-space'] * len(annot_ids))

        imgspace_canvas = dets_in_imgspace.draw_on(imgspace_canvas_raw.copy(), color='orange', labels=['img-space'] * len(annot_ids))
        imgspace_canvas = dets_in_vidspace.draw_on(imgspace_canvas, color='dodgerblue', labels=['vid-space'] * len(annot_ids))

        heatmask_in_vidspace = kwimage.ensure_alpha_channel(mask_in_vidspace, alpha=0.4)
        heatmask_in_imgspace = kwimage.ensure_alpha_channel(mask_in_imgspace, alpha=0.4)
        overlay_in_vidspace = kwimage.overlay_alpha_layers([heatmask_in_vidspace, vidspace_canvas_raw])
        overlay_in_imgspace = kwimage.overlay_alpha_layers([heatmask_in_imgspace, imgspace_canvas_raw])

        pnum_ = kwplot.PlotNums(nRows=3, nCols=2)
        kwplot.imshow(imgspace_canvas, fnum=1, pnum=pnum_(), title='img-space-canvas', show_ticks=1)
        kwplot.imshow(vidspace_canvas, fnum=1, pnum=pnum_(), title='vid-space-canvas', show_ticks=1)

        kwplot.imshow(mask_in_imgspace, fnum=1, pnum=pnum_(), title='img-space-mask', show_ticks=1)
        kwplot.imshow(mask_in_vidspace, fnum=1, pnum=pnum_(), title='vid-space-mask', show_ticks=1)

        kwplot.imshow(overlay_in_imgspace, fnum=1, pnum=pnum_(), title='img-space-overlay', show_ticks=1)
        kwplot.imshow(overlay_in_vidspace, fnum=1, pnum=pnum_(), title='vid-space-overlay', show_ticks=1)


def demo_change_pairs():
    # display with matplotlib
    DRAW = 1
    if DRAW:
        import kwplot
        kwplot.autompl()

    _default = ub.expandpath('$HOME/data/dvc-repos/smart_watch_dvc')
    dvc_dpath = os.environ.get('DVC_DPATH', _default)
    # coco_fpath = join(dvc_dpath, 'drop1-S2-L8-aligned/data.kwcoco.json')
    bundle_dpath = join(dvc_dpath, 'drop1-S2-L8-aligned')
    coco_fpath = join(bundle_dpath, 'combo_data.kwcoco.json')
    dset = kwcoco.CocoDataset(coco_fpath)

    # Create a sliding window object for each specific image (because they may
    # have different sizes, technically we could memoize this)
    import kwarray
    window_overlap = 0.5
    window_dims = (96, 96)
    keepbound = False
    vidid_to_space_slider = {}
    for vidid, video in dset.index.videos.items():
        full_dims = [video['height'], video['width']]
        window_dims_ = full_dims if window_dims == 'full' else window_dims
        slider = kwarray.SlidingWindow(full_dims, window_dims_,
                                       overlap=window_overlap,
                                       keepbound=keepbound,
                                       allow_overshoot=True)
        vidid_to_space_slider[vidid] = slider

    # from ndsampler import isect_indexer
    # _isect_index = isect_indexer.FrameIntersectionIndex.from_coco(dset)

    all_regions = []
    positive_idxs = []
    negative_idxs = []

    print('dset.cats = {}'.format(ub.urepr(dset.cats, nl=1)))

    # Given an video
    # video_id = ub.peek()
    for video_id in dset.index.videos.keys():
        slider = vidid_to_space_slider[video_id]

        video_info = dset.index.videos[video_id]
        video_gids = dset.index.vidid_to_gids[video_id]

        import pyqtree
        qtree = pyqtree.Index((0, 0, video_info['width'], video_info['height']))
        qtree.aid_to_ltrb = {}

        tid_to_info = ub.ddict(list)
        video_aids = dset.images(video_gids).annots.lookup('id')
        for aids, gid in zip(video_aids, video_gids):

            warp_vid_from_img = kwimage.Affine.coerce(dset.index.imgs[gid]['warp_img_to_vid'])
            # warp_img_from_vid = warp_vid_from_img.inv()

            img_info = dset.index.imgs[gid]
            frame_index = img_info['frame_index']
            tids = dset.annots(aids).lookup('track_id')
            cids = dset.annots(aids).lookup('category_id')
            for tid, aid, cid in zip(tids, aids, cids):

                imgspace_box = kwimage.Boxes([dset.index.anns[aid]['bbox']], 'xywh')
                vidspace_box = imgspace_box.warp(warp_vid_from_img)

                tlbr_box = vidspace_box.to_tlbr().data[0]
                qtree.insert(aid, tlbr_box)
                qtree.aid_to_ltrb[aid] = tlbr_box

                dset.index.anns[aid]['bbox']

                tid_to_info[tid].append({
                    'gid': gid,
                    'cid': cid,
                    'frame_index': frame_index,
                    # 'box': box,
                    'cname': dset._resolve_to_cat(cid)['name'],
                    'aid': aid,
                })

        # print('tid_to_info = {}'.format(ub.urepr(tid_to_info, nl=2, sort=0)))
        for space_region in list(slider):
            y_sl, x_sl = space_region

            # Find all annotations that pass through this spatial region
            vid_box = kwimage.Boxes.from_slice((y_sl, x_sl))
            query = vid_box.to_tlbr().data[0]
            isect_aids = sorted(set(qtree.intersect(query)))

            isect_annots = dset.annots(isect_aids)
            unique_tracks = set(isect_annots.lookup('track_id'))
            if len(unique_tracks) > 1:
                break

            frame_idxs = isect_annots.images.lookup('frame_index')
            isect_annots = isect_annots.take(ub.argsort(frame_idxs))

            frame_idxs = isect_annots.images.lookup('frame_index')
            track_ids = isect_annots.lookup('track_id')

            region_tid_to_aids = ub.group_items(isect_annots._ids, track_ids)
            region_tid_to_info = {}
            # For each track gather, start and stop info within this region
            for tid, aids in region_tid_to_aids.items():
                cids = dset.annots(aids).lookup('category_id')

                # phase_boundaries = np.diff(cids)

                frame_idxs = dset.annots(aids).images.lookup('frame_index')
                region_tid_to_info[tid] = {
                    'tid': tid,
                    'frame_idxs': frame_idxs,
                    'cids': cids,
                    'aids': aids,
                }
                pass

            pre_cid  = dset.index.name_to_cat['No Activity']['id']
            post_cid = dset.index.name_to_cat['Post Construction']['id']
            ignore_cid = dset.index.name_to_cat['Unknown']['id']

            import itertools as it
            video_frame_idxs = list(range(len(video_gids)))

            for frame_idxs in list(it.combinations(video_frame_idxs, 2)):

                gids = list(ub.take(video_gids, frame_idxs))

                has_change = False
                region_tracks = []
                for tid, track_info in region_tid_to_info.items():
                    sample_cids = []
                    for fx in frame_idxs:
                        idx = np.searchsorted(track_info['frame_idxs'], fx, 'right') - 1
                        if idx < 0:
                            cid = pre_cid
                        else:
                            cid = track_info['cids'][idx]
                        sample_cids.append(cid)

                    if len(sample_cids) and sample_cids[0] == post_cid:
                        # any leading "change completed" that transition into
                        # no-change labels should be marked as negative.
                        sample_cids
                        pass

                    if len(set(sample_cids) - {ignore_cid}) > 1:
                        # This frame sampling has a change in it
                        has_change = True

                    region_tracks.append({
                        'tid': tid,
                        'sample_cids': sample_cids,
                    })

                if has_change:
                    positive_idxs.append(len(all_regions))
                else:
                    negative_idxs.append(len(all_regions))

                all_regions.append({
                    'gids': gids,
                    'space_slice': space_region,
                    'has_change': has_change,
                    'region_tracks': region_tracks,
                })

    print(len(positive_idxs))
    print(len(negative_idxs))

    for idx in positive_idxs:
        print(all_regions[idx])



"""
Main tracker logic

SeeAlso:
    * ../../cli/run_tracker.py
"""
import ubelt as ub
import math
from typing import Tuple
from typing import Literal
import scriptconfig as scfg

from geowatch.heuristics import SITE_SUMMARY_CNAME, CNAMES_DCT
from geowatch.tasks.tracking.abstract_classes import NewTrackFunction

from geowatch.tasks.tracking.old_polygon_extraction import PolygonExtractConfig
from geowatch.tasks.tracking.old_polygon_extraction import _gids_polys, FoundNothing

from geowatch.tasks.tracking.utils import (
    _validate_keys,
    score_track_polys,
    trackid_is_default,
    gpd_sort_by_gid, gpd_len,
    gpd_compute_scores)

try:
    from line_profiler import profile
except Exception:
    profile = ub.identity

#
# --- track/polygon filters ---
#


class DataFrameFilter:

    def __call__(self, gdf):
        raise AssertionError("Use the explicit .filter_dataframe method instead")
        return self.filter_dataframe(gdf)

    def filter_dataframe(self, gdf):
        raise NotImplementedError


class TimePolygonFilter(DataFrameFilter):
    """
    Cuts off start and end of each track based on min response.
    """

    def __init__(self, threshold):
        self.threshold = threshold

    def filter_dataframe(self, gdf):

        def _edit(grp):
            magic_thresh = 0.5
            (ok_ixs, ) = (grp[('fg', self.threshold)] >
                          magic_thresh).values.nonzero()
            if len(ok_ixs) == 0:
                start_ix, end_ix = len(grp), len(grp)
            else:
                start_ix, end_ix = ok_ixs[[0, -1]]
            # print(grp.name, start_ix, end_ix+1)
            return grp.iloc[start_ix:end_ix + 1]

        if len(gdf) > 0:
            group = gdf.groupby('track_idx', group_keys=False)
            result = group.apply(_edit)
        else:
            result = gdf
        return result


class TimeSplitFilter(DataFrameFilter):
    """
    Splits tracks based on start and end of each subtracks min response.
    """

    def __init__(self, threshold, frame_buffer):
        self.threshold = threshold
        self.frame_buffer = frame_buffer

    def filter_dataframe(self, gdf):
        import geopandas as gpd
        import pandas as pd

        def buffer_by(tracks, by):
            new_tracks = []
            for start, end in tracks:
                new_tracks.append((start - by, end + by))
            return new_tracks

        def merge_neighbors(tracks):
            new_tracks = []
            prev = None
            for curr in tracks:
                if prev is None:
                    prev = curr
                    continue

                if curr[0] <= prev[1]:
                    prev = (prev[0], curr[1])
                else:
                    new_tracks.append(prev)
                    prev = curr

            new_tracks.append(prev)
            return new_tracks

        def _edit(scores):
            magic_thresh = 0.5
            track_start = None
            sub_tracks = []
            for idx, score in enumerate(scores):

                if (score > magic_thresh) and (track_start is None):
                    # print(f"track started at {idx}")
                    track_start = idx

                if (score < magic_thresh) and (track_start is not None):
                    # print(f"track ended at {idx-1}")
                    sub_tracks.append((track_start, idx))
                    track_start = None

            if (track_start is not None):
                sub_tracks.append((track_start, len(scores)))

            return sub_tracks

        if len(gdf) > 0:
            subtracks = []
            subtrack_idx = 1
            for track_id, group in gdf.groupby('track_idx'):

                subtrack_startstops = _edit(list(group[('fg', self.threshold)]))
                subtrack_startstops = buffer_by(subtrack_startstops, self.frame_buffer)
                subtrack_startstops = merge_neighbors(subtrack_startstops)

                if len(subtrack_startstops) == 0:
                    return gpd.GeoDataFrame()

                if subtrack_startstops[0][0] < 0:
                    subtrack_startstops[0] = (0, subtrack_startstops[0][1])
                if subtrack_startstops[-1][0] >= len(group):
                    subtrack_startstops[-1] = (subtrack_startstops[-1][0], len(group))

                for sub_id, (start, stop) in enumerate(subtrack_startstops):
                    subtrack = group.iloc[start:stop]
                    subtrack["track_idx"] = subtrack_idx
                    subtrack_idx += 1

                    subtracks.append(subtrack)
            result = gpd.GeoDataFrame(
                pd.concat(subtracks, ignore_index=True)
            )
        else:
            result = gdf
        return result


class ResponsePolygonFilter(DataFrameFilter):
    """
    Filters each track based on the average response of all tracks.
    """

    def __init__(self, gdf, threshold):

        self.threshold = threshold

        gids = gdf['gid'].unique()
        mean_response = gdf[('fg', -1)].mean()

        self.gids = gids
        self.mean_response = mean_response

    def filter_dataframe(self, gdf, gids=None, threshold=None, cross=True):
        if gids is None:
            gids = self.gids
        if threshold is None:
            threshold = self.threshold
        if cross:

            def _filter(grp):
                this_response = grp[grp['gid'].isin(self.gids)][('fg',
                                                                 -1)].mean()
                return this_response / self.mean_response > threshold

            return gdf.groupby('track_idx', group_keys=False).filter(_filter)
        else:
            cond = (gdf[('fg', -1)] / self.mean_response > threshold)
            return gdf[cond]


#
# --- main logic ---
#


@profile
def _add_tracks_to_dset(sub_dset, tracks, thresh, key, bg_key=None):
    """
    This takes the GeoDataFrame with computed or modified tracks and adds them
    to ``sub_dset``.

    We are assuming the polygon geometry in "tracks" is in video space.
    """
    import kwcoco
    import kwimage
    from geowatch.utils import kwcoco_extensions
    key, bg_key = _validate_keys(key, bg_key)

    print('Add tracks to dset')
    print(f'bg_key={bg_key}')
    print(f'key={key}')
    # print('tracks:')
    # print(tracks)

    if tracks.empty:
        print('no tracks to add!')
        return sub_dset

    @ub.memoize
    def _warp_img_from_vid(gid):
        # Memoize the conversion to a matrix
        coco_img = sub_dset.coco_image(gid)
        img_from_vid = coco_img.warp_img_from_vid
        return img_from_vid

    def make_new_annotation(gid, poly, this_score, scores_dct, track_id,
                            space='video'):

        # assign category (key) from max score
        if this_score > thresh or len(bg_key) == 0:
            cand_keys = key
        else:
            cand_keys = bg_key
        if 1:
            # HACK for eval16, need to be nicer about what we do here
            if len(cand_keys) > 1:
                cand_keys = ub.oset(cand_keys) - {'ac_salient'}
        if len(cand_keys) > 1:
            # TODO ensure bg classes are scored if there are >1 of them
            cat_name = ub.argmax(ub.udict.subdict(scores_dct, cand_keys))
        else:
            cat_name = cand_keys[0]
        cid = sub_dset.ensure_category(cat_name)

        assert space in {'image', 'video'}
        if space == 'video':
            # Transform the video polygon into image space
            img_from_vid = _warp_img_from_vid(gid)
            poly = kwimage.MultiPolygon.coerce(poly).warp(img_from_vid)

        bbox = list(poly.box().boxes.to_coco())[0]
        segmentation = poly.to_coco(style='new')

        # Add the polygon as an annotation on the image
        new_ann = dict(image_id=gid,
                       category_id=cid,
                       bbox=bbox,
                       segmentation=segmentation,
                       score=this_score,
                       scores=scores_dct,
                       track_id=track_id)
        return new_ann

    new_trackids = kwcoco_extensions.TrackidGenerator(sub_dset)

    all_new_anns = []

    def _add(obs, tid):
        if not trackid_is_default(tid):
            track_id = tid
            new_trackids.exclude_trackids([track_id])
        else:
            track_id = next(new_trackids)

        for o in obs:
            new_ann = make_new_annotation(*o, track_id)
            all_new_anns.append(new_ann)

    try:
        tracks.groupby('track_idx', axis=0)
    except ValueError:
        import warnings
        warnings.warn('warning: no tracks to add the the kwcoco dataset')
    else:
        # ASSIGN_ANNOTS_SCORE
        score_chan = kwcoco.ChannelSpec('|'.join(key))

        for tid, grp in tracks.groupby('track_idx', axis=0):

            try:
                this_score = grp[(score_chan.spec, -1)]
            except Exception:
                # HACK
                this_score = grp[('ac_salient', -1)]

            scores_dct = {k: grp[(k, -1)] for k in score_chan.unique()}
            scores_dct = [dict(zip(scores_dct, t))
                          for t in zip(*scores_dct.values())]
            _obs_iter = zip(grp['gid'], grp['poly'], this_score, scores_dct)
            _add(_obs_iter, tid)

    # TODO: Faster to add annotations in bulk, but we need to construct the
    # "ids" first
    for new_ann in all_new_anns:
        sub_dset.add_annotation(**new_ann)

    DEBUG_JSON_SERIALIZABLE = 0
    if DEBUG_JSON_SERIALIZABLE:
        from kwutil.util_json import debug_json_unserializable
        debug_json_unserializable(sub_dset.dataset)

    return sub_dset


@profile
def site_validation(sub_dset, thresh=0.25, span_steps=15):
    """
    Example:
        >>> import geowatch
        >>> from geowatch.tasks.tracking.from_heatmap import *  # NOQA
        >>> coco_dset = geowatch.coerce_kwcoco(
        >>>     'geowatch-msi', heatmap=True, geodata=True, dates=True)
        >>> vid_id = coco_dset.videos()[0]
        >>> sub_dset = coco_dset.subset(list(coco_dset.images(video_id=vid_id)))
        >>> import numpy as np
        >>> for ann in sub_dset.anns.values():
        >>>     ann['score'] = float(np.random.rand())
        >>> sub_dset.remove_annotations(sub_dset.index.trackid_to_aids[None])
        >>> sub_dset = site_validation(sub_dset)
    """

    # Turn annotations into table we can query
    # annots = pd.DataFrame([
    #     (ub.udict(ann) & {'score', 'track_id', 'track_idx'})
    #     # {
    #     #     'score': ann['score'],
    #     #     'track_id': ann.get('track_id', None),
    #     #     'track_idx': ann.get('track_idx', None),
    #     # }
    #     for ann in sub_dset.dataset["annotations"]
    # ])
    import pandas as pd
    imgs = pd.DataFrame(sub_dset.dataset['images'])
    if 'timestamp' not in imgs.columns:
        imgs['timestamp'] = imgs['id']

    annots = pd.DataFrame(sub_dset.dataset['annotations'])

    if annots.shape[0] == 0:
        print('Nothing to filter')
        return sub_dset

    annots = annots[[
        'id', 'image_id', 'track_id', 'score'
    ]].join(
        imgs[['timestamp']],
        on='image_id',
    )

    track_ids_to_drop = []
    ann_ids_to_drop = []

    for track_id, track_group in annots.groupby('track_id', axis=0):

        # Scores are inherently noisy. We smooth them out with a
        # `span_steps`-wide weighted moving average. The maximum
        # value of this decides whether to keep the track.
        # TODO: do something more elegant here?
        score = track_group['score'].ewm(span=span_steps).mean().max()
        if score < thresh:
            track_ids_to_drop.append(track_id)
            ann_ids_to_drop.extend(track_group['id'].tolist())

    print(f"Dropping {len(ann_ids_to_drop)} annotations from {len(track_ids_to_drop)} tracks.")
    if len(ann_ids_to_drop) > 0:
        sub_dset.remove_annotations(ann_ids_to_drop)

    return sub_dset


@profile
def time_aggregated_polys(sub_dset, video_id, **kwargs):
    """
    Polygon extraction and tracking function.

    Aggregate heatmaps across time, threshold them to get polygons,
    and add one track per polygon.

    Args:
        sub_dset (kwcoco.CocoDataset): a kwcoco dataset with exactly 1 video

        video_id (int): The video-id to track.

        **kwargs:
            see :class:`TimeAggregatedPolysConfig` and
            :class:`PolygonExtractConfig`.

    Ignore:
        # For debugging
        import xdev
        from geowatch.tasks.tracking.from_heatmap import *  # NOQA
        from geowatch.tasks.tracking.from_heatmap import _validate_keys
        globals().update(xdev.get_func_kwargs(time_aggregated_polys))

    Example:
        >>> # test interpolation
        >>> from geowatch.tasks.tracking.from_heatmap import time_aggregated_polys
        >>> from geowatch.demo import demo_kwcoco_with_heatmaps
        >>> import geowatch
        >>> sub_dset = geowatch.coerce_kwcoco(
        >>>     'geowatch-msi', num_videos=1, num_frames=5, image_size=(128, 128),
        >>>     geodata=True, heatmap=True, dates=True)
        >>> thresh = 0.01
        >>> video_id = list(sub_dset.videos())[0]
        >>> min_area_square_meters = None
        >>> kwargs = dict(thresh=thresh, min_area_square_meters=min_area_square_meters, time_thresh=None)
        >>> orig_track = time_aggregated_polys(sub_dset, video_id, **kwargs)
        >>> # Test robustness to frames that are missing heatmaps
        >>> skip_gids = [1,3]
        >>> for gid in skip_gids:
        >>>      sub_dset.imgs[gid]['auxiliary'].pop()
        >>> inter_track = time_aggregated_polys(sub_dset, video_id, **kwargs)
        >>> assert inter_track.iloc[0][('fg', -1)] == 0
        >>> assert inter_track.iloc[1][('fg', -1)] > 0

    Example:
        >>> # test interpolation
        >>> from geowatch.tasks.tracking.from_heatmap import time_aggregated_polys
        >>> from geowatch.demo import demo_kwcoco_with_heatmaps
        >>> import geowatch
        >>> sub_dset = geowatch.coerce_kwcoco(
        >>>     'geowatch-msi', num_videos=1, num_frames=5, image_size=(128, 128),
        >>>     geodata=True, heatmap=True, dates=True)
        >>> video_id = list(sub_dset.videos())[0]
        >>> thresh = 0.01
        >>> min_area_square_meters = None
        >>> kwargs = dict(thresh=thresh, min_area_square_meters=min_area_square_meters, time_thresh=None)
        >>> orig_track = time_aggregated_polys(sub_dset, video_id, **kwargs)
        >>> # Test robustness to frames that are missing heatmaps
        >>> skip_gids = [1,3]
        >>> for gid in skip_gids:
        >>>      sub_dset.imgs[gid]['auxiliary'].pop()
        >>> inter_track = time_aggregated_polys(sub_dset, video_id, **kwargs)
        >>> assert inter_track.iloc[0][('fg', -1)] == 0
        >>> assert inter_track.iloc[1][('fg', -1)] > 0
    """
    #
    # --- input validation ---
    #
    import kwimage
    import geopandas as gpd
    import rich
    config = TimeAggregatedPolysConfig(**kwargs)
    config.key, config.bg_key = _validate_keys(config.key, config.bg_key)

    _all_keys = set(config.key + config.bg_key)
    has_requested_chans_list = []

    # coco_videos = sub_dset.videos()
    # assert len(coco_videos) == 1, 'we expect EXACTLY one video here'
    assert video_id is not None
    coco_videos = sub_dset.videos(video_ids=[video_id])
    video = coco_videos.objs[0]
    video_name = video.get('name', None)
    # video_id = video['id']

    video_gids = list(sub_dset.images(video_id=video_id))
    for gid in video_gids:
        coco_img = sub_dset.coco_image(gid)
        chan_codes = coco_img.channels.normalize().fuse().as_set()
        flag = bool(_all_keys & chan_codes)
        has_requested_chans_list.append(flag)

    scale_vid_from_trk, tracking_gsd = _determine_tracking_scale(
        config, sub_dset, video_gids, video)

    if not any(has_requested_chans_list):
        raise KeyError(f'no imgs in dset {sub_dset.tag} '
                       f'have keys {config.key} or {config.bg_key}.')
    if not all(has_requested_chans_list):
        n_total = len(has_requested_chans_list)
        n_have = sum(has_requested_chans_list)
        n_missing = (n_total - n_have)
        print(f'warning: {n_missing} / {n_total} imgs in dset {sub_dset.tag} '
              f'with video {video_name} have no keys {config.key} or {config.bg_key}. '
              'Interpolating...')

    #
    # --- main logic ---
    #

    # polys are in "tracking-space", i.e. video-space up to a scale factor.
    gid_poly_config = PolygonExtractConfig(**ub.udict(config).subdict(PolygonExtractConfig.__default__.keys()))
    try:
        gids_polys = _gids_polys(sub_dset, video_id, **gid_poly_config)
    except FoundNothing:
        gids_polys = []

    orig_gid_polys = list(gids_polys)  # 26% of runtime
    gids_polys = orig_gid_polys

    if len(gids_polys):
        rich.print('[green] time aggregation: number of polygons: ', len(gids_polys))
    else:
        rich.print('[red] time aggregation: number of polygons: ', len(gids_polys))

    # size and response filters should operate on each vidpoly separately.
    if config.max_area_square_meters:
        max_area_sqpx = config.max_area_square_meters / (tracking_gsd ** 2)
        n_orig = len(gids_polys)
        if config.max_area_behavior == 'drop':
            gids_polys = [(t, p) for t, p in gids_polys
                          if p.to_shapely().area < max_area_sqpx]
            print('filter large: remaining polygons: '
                  f'{len(gids_polys)} / {n_orig}')
        elif config.max_area_behavior == 'grid':
            # edits tracks instead of removing them
            raise NotImplementedError(config.max_area_behavior)
        elif config.max_area_behavior == 'ignore':
            # Do nothing, just let it go through.
            ...
        else:
            raise KeyError(config.max_area_behavior)

    if config.min_area_square_meters:
        min_area_sqpx = config.min_area_square_meters / (tracking_gsd ** 2)
        n_orig = len(gids_polys)
        gids_polys = [(t, p) for t, p in gids_polys
                      if p.to_shapely().area > min_area_sqpx]
        print('filter small: remaining polygons: '
              f'{len(gids_polys)} / {n_orig}')

    # now we start needing scores, so bulk-compute them

    gids_polys_T = list(zip(*gids_polys))
    if gids_polys_T:
        gids, polys = gids_polys_T
    else:
        gids, polys = [], []

    polys = [p.to_shapely() for p in polys]

    # At this point each row corresponds to a single track and the each gid
    # cell contains a list of image ids.
    _TRACKS_COMPACT = gpd.GeoDataFrame({'gid': gids, 'poly': polys}, geometry='poly')
    if config.polygon_simplify_tolerance is not None:
        _TRACKS_COMPACT['poly'] = _TRACKS_COMPACT['poly'].simplify(tolerance=config.polygon_simplify_tolerance)
    _TRACKS_COMPACT = _TRACKS_COMPACT.reset_index(names='track_idx')

    # Explode takes each row with multiple gids and expands it creating a new
    # row for each item in the exploeded column. That means we go from a
    # dataframe that looks like:
    # [
    #   {'gid': [1, 2, 3], 'track_idx': 0, 'poly': POLY1},
    #   {'gid': [5, 7], 'track_idx': 1, 'poly': POLY2},
    # ]
    # TO:
    # [
    #   {'gid': 1, 'track_idx': 0, 'poly': POLY1},
    #   {'gid': 2, 'track_idx': 0, 'poly': POLY1},
    #   {'gid': 3, 'track_idx': 0, 'poly': POLY1},
    #   {'gid': 5, 'track_idx': 1, 'poly': POLY2},
    #   {'gid': 7, 'track_idx': 1, 'poly': POLY2},
    # ]
    _TRACKS = _TRACKS_COMPACT.explode('gid')

    # ensure index is sorted in video order
    sorted_gids = sub_dset.images(video_id=video['id']).gids
    _TRACKS = gpd_sort_by_gid(_TRACKS, sorted_gids)

    # awk, find better way of bookkeeping and indexing into scores needed
    thrs = {-1}
    if config.response_thresh:
        thrs.add(-1)
    if config.time_thresh:
        thrs.add(config.time_thresh * config.thresh)
    if config.time_split_thresh:
        thrs.add(config.time_split_thresh)
    #####
    ## Jon C: I'm not sure about this. Going from a set to a list, and then having
    ## the resulting function depend on the order of the list makes me nerevous.
    #####
    thrs = list(thrs)

    """
    Cases:
        For BAS:
            ks = {'fg': ['salient'], 'bg': []}

        For AC:
            {'fg': ['Site Preparation', 'Active Construction', 'Post Construction', 'No Activity', 'ac_salient'],
             'bg': ['No Activity']}

    """

    ks = {'fg': config.key, 'bg': config.bg_key}

    # TODO dask gives different results on polys that overlap nodata area, need
    # to debug this. (6% of polygons in KR_R001, so not a huge difference)
    # USE_DASK = True
    USE_DASK = False
    print('Begin compute track scores:')
    # Note: this is the function also called by
    # :func:`score_track_polys`

    modulate = None
    if config.modulate_post_construction is not None:
        modulate = {}
        modulate['Post Construction'] = float(config.modulate_post_construction)

    _TRACKS = gpd_compute_scores(_TRACKS, sub_dset, thrs, ks,
                                 USE_DASK=USE_DASK,
                                 resolution=config.resolution,
                                 modulate=modulate)

    rich.print('[green]Finished computing track scores:')
    # rich.print(_TRACKS)
    if _TRACKS.empty:
        return _TRACKS

    # dask could unsort
    _TRACKS = gpd_sort_by_gid(_TRACKS.reset_index(), sorted_gids)

    # response_thresh = 0.9
    if config.response_thresh:

        n_orig = gpd_len(_TRACKS)
        rsp_filter = ResponsePolygonFilter(_TRACKS, config.key, config.response_thresh)
        _TRACKS = rsp_filter.filter_dataframe(_TRACKS)
        print('filter based on per-polygon response: remaining tracks '
              f'{gpd_len(_TRACKS)} / {n_orig}')

    # TimePolygonFilter edits tracks instead of removing them
    if config.time_thresh:  # as a fraction of thresh
        time_filter = TimePolygonFilter(config.time_thresh * config.thresh)
        n_orig = gpd_len(_TRACKS)
        _TRACKS = time_filter.filter_dataframe(_TRACKS)  # 7% of runtime? could be next line
        print('filter based on time overlap: remaining tracks '
              f'{gpd_len(_TRACKS)} / {n_orig}')

    if config.time_split_thresh:
        split_filter = TimeSplitFilter(config.time_split_thresh, config.time_split_frame_buffer)
        n_orig = gpd_len(_TRACKS)
        _TRACKS = split_filter.filter_dataframe(_TRACKS)
        n_result = gpd_len(_TRACKS)
        print('filter based on time splitting: remaining tracks '
              f'{n_result} / {n_orig}')

    # The tracker assumes the polygons will be output in video space.
    # rich.print('[red]!!!!!!!!!')
    # print(f'scale_vid_from_trk={scale_vid_from_trk}')
    # print(f'scale_vid_from_trk={scale_vid_from_trk}')
    # print(f'scale_vid_from_trk={scale_vid_from_trk}')
    # print(f'scale_vid_from_trk={scale_vid_from_trk}')
    # rich.print('[red]!!!!!!!!!')
    if scale_vid_from_trk is not None and len(_TRACKS):
        # If a tracking resolution was specified undo the extra scale factor
        _TRACKS['poly'] = _TRACKS['poly'].scale(*scale_vid_from_trk, origin=(0, 0))

    # TODO: do we need to convert to MultiPolygon here? Or can that be handled
    # by consumers of this method?
    _TRACKS['poly'] = _TRACKS['poly'].map(kwimage.MultiPolygon.from_shapely)
    rich.print('[green]Returning Tracks')
    # rich.print(_TRACKS)
    return _TRACKS


def _determine_tracking_scale(config, sub_dset, video_gids, video):
    """
    Factored out code from :func:`time_aggregated_polys`
    """
    import numpy as np
    scale_vid_from_trk = None
    tracking_gsd = None
    if len(video_gids) and (config.resolution is not None):
        # Determine resolution information for videospace (what we will return
        # in) and tracking space (what we will build heatmaps in)
        first_gid = video_gids[0]
        first_coco_img = sub_dset.coco_image(first_gid)
        # (w, h)
        vidspace_resolution = first_coco_img.resolution(space='video')['mag']
        vidspace_resolution = np.array(vidspace_resolution)

        # (w, h)
        scale_trk_from_vid = first_coco_img._scalefactor_for_resolution(
            space='video', resolution=config.resolution)
        scale_trk_from_vid = np.array(scale_trk_from_vid)

        # Determinethe pixel size of tracking space
        tracking_resolution = vidspace_resolution / scale_trk_from_vid
        if not np.isclose(*tracking_resolution):
            print(f'warning: nonsquare pxl size of {tracking_resolution}')
        tracking_gsd = np.mean(tracking_resolution)

        # Get the transform from tracking space back to video space
        scale_vid_from_trk = 1 / scale_trk_from_vid
    else:
        scale_vid_from_trk = (1, 1)

    if tracking_gsd is None:
        if len(video_gids):
            # Use whatever is in the kwcoco file as the default.
            first_gid = video_gids[0]
            first_coco_img = sub_dset.coco_image(first_gid)
            # (w, h)
            try:
                vidspace_resolution = first_coco_img.resolution(space='video')['mag']
                default_gsd = np.mean(vidspace_resolution)
            except Exception:
                default_gsd = None
        else:
            default_gsd = 30
            print(f'warning: video {video["name"]} in dset {sub_dset.tag} '
                  f'has no listed resolution; assuming {default_gsd}')
        tracking_gsd = default_gsd
    return scale_vid_from_trk, tracking_gsd


#
# --- wrappers ---
#
# Note:
#     The following are valid choices of `track_fn` in
#     ../../cli/run_tracker.py and will be called by ./normalize.py

class TimeAggregatedPolysConfig(PolygonExtractConfig):
    """
    This is an intermediate config that we will use to transition between the
    current dataclass configuration and a new scriptconfig based one.

    python -c "if 1:
        from geowatch.tasks.tracking.from_heatmap import TimeAggregatedBAS
        TimeAggregatedBAS().argparse().print_help()
    "

    """
    bg_key = scfg.Value(None, help=ub.paragraph(
        '''
        Zero or more channels to use as the negative class for polygon scoring.

        bg_key (String | List[String] | None): background key(s).
            If None, background heatmaps become 1 - sum(foreground keys)
        '''))

    time_split_thresh = scfg.Value(None, help=ub.paragraph(
        '''
        time splitting parameter. if set, tracks will be broken into subtracks
        based on when the score is above this threshold.
        '''))

    time_split_frame_buffer = scfg.Value(2, help=ub.paragraph(
        '''
        time splitting parameter. if set, subtracks will be buffered by the specified
        number of frames. if this causes subtracks to overlap, they are merged together.
        '''))

    time_thresh = scfg.Value(1, help=ub.paragraph(
        '''
        Multiplier on the regular threshold used to determine the temporal
        extent of the polygon over time. All polygons must have an aggregate
        score over ``thresh * time_thresh``.  Typically set this a bit less
        than 1. (e.g. 0.8).
        '''))

    response_thresh = scfg.Value(None, help=ub.paragraph(
        '''
        I dont remember what this does. Help wanted with documenting.
        '''))

    min_area_square_meters = scfg.Value(None, help=ub.paragraph(
        '''
        If specified, any site with an area less than this threshold is
        removed.
        '''))

    max_area_square_meters = scfg.Value(None, help=ub.paragraph(
        '''
        If specified, any site with an area greater than this threshold is
        removed.
        '''))

    max_area_behavior = scfg.Value('drop', help=ub.paragraph(
        '''
        How to handle polygons that are over the max area threshold.
        '''))

    polygon_simplify_tolerance = scfg.Value(None, help=ub.paragraph(
        '''
        The pixel size (at the specified heatmap resolution) to use for polygon
        simplification.
        '''))

    modulate_post_construction = scfg.Value(None, help=ub.paragraph(
        '''
        Hacked in POC command to multiply post scores.
        '''))

    def __post_init__(self):
        super().__post_init__()
        if self.norm_ord in {'inf', None}:
            self.norm_ord = float('inf')
        # self.key, self.bg_key = _validate_keys(self.key, self.bg_key)

        if isinstance(self.inner_window_size, float) and math.isnan(self.inner_window_size):
            self.inner_window_size = None

        if isinstance(self.moving_window_size, float) and math.isnan(self.moving_window_size):
            self.moving_window_size = None


class CommonTrackFn(NewTrackFunction, TimeAggregatedPolysConfig):
    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.norm_ord, str) and self.norm_ord.lower() == 'inf':
            self.norm_ord = float('inf')


class TrackFnWithSV(CommonTrackFn):
    site_validation: bool = False
    site_validation_span_steps: int = 120
    site_validation_thresh: float = 0.1


class TimeAggregatedBAS(TrackFnWithSV):
    """
    Wrapper for BAS that looks for change heatmaps.
    """
    thresh: float = 0.2
    key: str = 'salient'
    agg_fn: str = 'probs'

    def create_tracks(self, sub_dset, video_id):
        aggkw = ub.udict(self) & TimeAggregatedPolysConfig.__default__.keys()
        tracks = time_aggregated_polys(sub_dset, video_id, **aggkw)
        # print('Tracks:')
        # print(tracks)
        return tracks

    def add_tracks_to_dset(self, sub_dset, tracks):
        sub_dset = _add_tracks_to_dset(sub_dset, tracks, self.thresh, self.key)
        if self.site_validation:
            sub_dset = site_validation(
                sub_dset,
                thresh=self.site_validation_thresh,
                span_steps=self.site_validation_span_steps,
            )
        return sub_dset


class TimeAggregatedSC(TrackFnWithSV):
    """
    Wrapper for Activity Characterization / Site Characterization that looks
    for phase heatmaps.

    Alias: class_heatmaps

    Note:
        This is a valid choice of `track_fn` in ../../cli/run_tracker.py
    """
    thresh: float = 0.01

    # key: Tuple[str] = tuple(CNAMES_DCT['positive']['scored'])

    # HACK TO REMEMBER ALL SCORES
    # TODO: Ensure this does not  break anything and refactor such that the
    # default behavior is to aggreate the score from all available classes when
    # only scoring. When refining polygons, a different approach is needed.
    key: Tuple[str] = tuple(['Site Preparation', 'Active Construction', 'Post Construction', 'No Activity', 'ac_salient'])

    # IS THIS USED?
    bg_key: Tuple[str] = tuple(CNAMES_DCT['negative']['scored'])

    boundaries_as: Literal['bounds', 'polys', 'none'] = 'bounds'
    time_thresh = None

    def create_tracks(self, sub_dset, video_id):
        """
        boundaries_as: use for Site Boundary annots in coco_dsennjk
            'bounds': generated polys will lie inside the boundaries
            'polys': generated polys will be the boundaries
            'none': generated polys will ignore the boundaries
        """
        import kwcoco
        import kwimage
        import rich
        rich.print('[white] --- Create Tracks ---')
        if self.boundaries_as == 'polys':

            # Just score the polygons, no need to extract
            tracks = score_track_polys(
                sub_dset,
                video_id,
                cnames=[SITE_SUMMARY_CNAME],
                # these are SC scores, not BAS, so this is not a
                # true reproduction of hybrid.
                score_chan=kwcoco.ChannelSpec('|'.join(self.key)),
                resolution=self.resolution,
            )
            # hack in always-foreground instead
            # tracks[(score_chan, -1)] = 1

            # try to ignore this error
            tracks['poly'] = tracks['poly'].map(
                kwimage.MultiPolygon.from_shapely)

        else:
            # Need to extract and score
            aggkw = ub.udict(self) & TimeAggregatedPolysConfig.__default__.keys()
            aggkw['use_boundaries'] = str(self.get('boundaries_as', 'none')).lower() not in {'none', 'null'}
            tracks = time_aggregated_polys(sub_dset, video_id, **aggkw)
        # print('Tracks:')
        # print(tracks)
        rich.print('[white] ---')
        return tracks

    def add_tracks_to_dset(self, sub_dset, tracks, **kwargs):
        import rich
        rich.print('[white] --- Add Tracks To Dataset ---')
        import kwcoco
        if self.boundaries_as != 'polys':
            if 0:
                col_map = {}
                for c in tracks.columns:
                    if c[0] == 'fg':
                        k = kwcoco.ChannelSpec('|'.join(self.key)).spec
                        col_map[c] = (k, *c[1:])
                    elif c[0] == 'bg':
                        k = kwcoco.ChannelSpec('|'.join(self.bg_key)).spec
                        col_map[c] = (k, *c[1:])
                print(f'col_map={col_map}')
                # weird effect here - reassignment casts from GeoDataFrame to
                # DataFrame. Related to invalid geometry column?
                # tracks = tracks.rename(columns=col_map)
                tracks.rename(columns=col_map, inplace=True)

        thresh = self.thresh
        key = self.key
        bg_key = self.bg_key
        # print(tracks)
        sub_dset = _add_tracks_to_dset(sub_dset, tracks=tracks, thresh=thresh,
                                       key=key, bg_key=bg_key, **kwargs)
        if self.site_validation:
            sub_dset = site_validation(
                sub_dset,
                thresh=self.site_validation_thresh,
                span_steps=self.site_validation_span_steps,
            )
        rich.print('[white] ---')
        return sub_dset


class TimeAggregatedSV(CommonTrackFn):
    """
    Wrapper for Site Validation that looks for phase heatmaps.

    Alias:
        site_validation

    Note:
        This is a valid choice of `track_fn` in ../../cli/run_tracker.py
    """
    thresh: float = 0.1
    key: str = 'salient'
    boundaries_as: Literal['bounds', 'polys', 'none'] = 'polys'
    span_steps: int = 120

    def create_tracks(self, sub_dset, video_id):
        """
        boundaries_as: use for Site Boundary annots in coco_dset
            'bounds': generated polys will lie inside the boundaries
            'polys': generated polys will be the boundaries
            'none': generated polys will ignore the boundaries
        """
        import kwcoco
        import kwimage
        if self.boundaries_as == 'polys':
            tracks = score_track_polys(
                sub_dset,
                video_id,
                cnames=[SITE_SUMMARY_CNAME],
                # these are SC scores, not BAS, so this is not a
                # true reproduction of hybrid.
                score_chan=kwcoco.ChannelSpec('|'.join((self.key,))),
                resolution=self.resolution,
            )
            # hack in always-foreground instead
            # tracks[(score_chan, None)] = 1

            # try to ignore this error
            tracks['poly'] = tracks['poly'].map(
                kwimage.MultiPolygon.from_shapely)

        else:
            raise NotImplementedError
        return tracks

    def add_tracks_to_dset(self, sub_dset, tracks, **kwargs):
        # if self.boundaries_as != 'polys':
        #     col_map = {}
        #     for c in tracks.columns:
        #         if c[0] == 'fg':
        #             k = kwcoco.ChannelSpec('|'.join(self.key)).spec
        #             col_map[c] = (k, *c[1:])
        #         elif c[0] == 'bg':
        #             k = kwcoco.ChannelSpec('|'.join(self.bg_key)).spec
        #             col_map[c] = (k, *c[1:])
        #     # weird effect here - reassignment casts from GeoDataFrame to
        #     # DataFrame. Related to invalid geometry column?
        #     # tracks = tracks.rename(columns=col_map)
        #     tracks.rename(columns=col_map, inplace=True)

        sub_dset = _add_tracks_to_dset(sub_dset, tracks, self.thresh, self.key,
                                       **kwargs)
        sub_dset = site_validation(
            sub_dset,
            thresh=self.thresh,
            span_steps=self.span_steps,
        )

        return sub_dset



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements building blocks for neural networks."""
from typing import Optional

import tensorflow as tf
from absl import logging
from deeplab2.model import utils
from deeplab2.model.layers import convolutions
from deeplab2.model.layers import squeeze_and_excite

backend = tf.keras.backend
layers = tf.keras.layers


class InvertedBottleneckBlock(tf.keras.layers.Layer):
    """An inverted bottleneck block.

    Reference:
    Sandler, M., Howard, A., et al. Mobilenetv2: Inverted residuals and linear
      bottlenecks. In CVPR, 2018
    Howard, A., Sandler, M., et al. Searching for mobilenetv3. In ICCV, 2019
    """

    def __init__(self,
                 in_filters: int,
                 out_filters: int,
                 expand_ratio: int,
                 strides: int,
                 kernel_size: int = 3,
                 se_ratio: Optional[float] = None,
                 activation: str = 'relu',
                 se_inner_activation: str = 'relu',
                 se_gating_activation: str = 'sigmoid',
                 depthwise_activation: Optional[str] = None,
                 expand_se_in_filters: bool = False,
                 atrous_rate: int = 1,
                 divisible_by: int = 1,
                 bn_layer: layers.Layer = tf.keras.layers.BatchNormalization,
                 conv_kernel_weight_decay: float = 0.0,
                 regularize_depthwise: bool = False,
                 use_depthwise: bool = True,
                 use_residual: bool = True,
                 name: Optional[str] = None):  # pytype: disable=annotation-type-mismatch  # typed-keras
        """Initializes an inverted bottleneck block with BN after convolutions.

        Args:
          in_filters: The number of filters of the input tensor.
          out_filters: The number of filters of the output tensor.
          expand_ratio: The expand_ratio for an inverted bottleneck block. If
            expand_ratio is <= 1, this argument will be ignored.
          strides: The number of stride. If greater than 1, this block will
            ultimately downsample the input.
          kernel_size: The kernel size of the depthwise conv layer.
          se_ratio: If not None, se ratio for the squeeze and excitation layer.
          activation: The name of the activation function.
          se_inner_activation: The name of squeeze-excitation inner activation.
          se_gating_activation: The name of squeeze-excitation gating activation.
          depthwise_activation: The name of the activation function for depthwise
            only.
          expand_se_in_filters: Whether or not to expand in_filter in squeeze and
            excitation layer.
          atrous_rate: The atrous dilation rate to use for.
          divisible_by: A number that all inner dimensions are divisible by.
          bn_layer: An optional tf.keras.layers.Layer that computes the
              normalization (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: The weight decay for convolution kernels.
          regularize_depthwise: Whether or not apply regularization on depthwise.
          use_depthwise: Whether to uses standard convolutions instead of depthwise.
          use_residual: Whether to include residual connection between input and
            output.
          name: Name for the block.
        """
        super(InvertedBottleneckBlock, self).__init__(name=name)

        self._in_filters = in_filters
        self._out_filters = out_filters
        self._expand_ratio = expand_ratio
        self._strides = strides
        self._kernel_size = kernel_size
        self._se_ratio = se_ratio
        self._divisible_by = divisible_by
        self._atrous_rate = atrous_rate
        self._regularize_depthwise = regularize_depthwise
        self._use_depthwise = use_depthwise
        self._use_residual = use_residual
        self._activation = activation
        self._se_inner_activation = se_inner_activation
        self._se_gating_activation = se_gating_activation
        self._depthwise_activation = depthwise_activation
        self._expand_se_in_filters = expand_se_in_filters

        if tf.keras.backend.image_data_format() == 'channels_last':
            self._bn_axis = -1
        else:
            self._bn_axis = 1

        if depthwise_activation is None:
            self._depthwise_activation = activation

        # if regularize_depthwise:
        #    depthwise_kernel_weight_decay = conv_kernel_weight_decay
        # else:
        #    depthwise_kernel_weight_decay = 0.0

        if self._expand_ratio <= 1 and not self._use_depthwise:
            raise ValueError(
                'Undefined behavior if expand_ratio <= 1 and not use_depthwise')

        expand_filters = self._in_filters
        if self._expand_ratio > 1:
            # First 1x1 conv for channel expansion.
            expand_filters = utils.make_divisible(
                self._in_filters * self._expand_ratio, self._divisible_by)

            expand_kernel = 1 if self._use_depthwise else self._kernel_size
            expand_stride = 1 if self._use_depthwise else self._strides

            self._conv1_bn_act = convolutions.Conv2DSame(
                output_channels=expand_filters,
                kernel_size=expand_kernel,
                strides=expand_stride,
                atrous_rate=1,
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation=self._activation,
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                name='expand_conv')

        if self._use_depthwise:
            # Depthwise conv.
            self._conv2_bn_act = convolutions.DepthwiseConv2DSame(
                kernel_size=self._kernel_size,
                strides=self._strides,
                atrous_rate=self._atrous_rate,
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation=self._depthwise_activation,
                name='depthwise_conv')

        # Squeeze and excitation.
        if self._se_ratio is not None and self._se_ratio > 0:
            if self._expand_se_in_filters:
                in_filters = expand_filters
            else:
                in_filters = self._in_filters
            self._squeeze_excitation = squeeze_and_excite.SqueezeAndExcite(
                in_filters=in_filters,
                out_filters=expand_filters,
                se_ratio=self._se_ratio,
                divisible_by=self._divisible_by,
                kernel_initializer='he_normal',
                kernel_regularizer=tf.keras.regularizers.l2(conv_kernel_weight_decay),
                activation=self._se_inner_activation,
                gating_activation=self._se_gating_activation,
                name=name + '_se')
        else:
            logging.info(
                'Squeeze and Excitation is skipped due to undefined se_ratio')
            self._squeeze_excitation = None

        # Last 1x1 conv.
        self._conv3_bn = convolutions.Conv2DSame(
            output_channels=self._out_filters,
            kernel_size=1,
            strides=1,
            atrous_rate=1,
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=None,
            conv_kernel_weight_decay=conv_kernel_weight_decay,
            name='project_conv')

    def call(self, inputs, training=None):
        depthwise_output = None
        shortcut = inputs
        if self._expand_ratio > 1:
            x = self._conv1_bn_act(inputs, training=training)
        else:
            x = inputs

        if self._use_depthwise:
            x = self._conv2_bn_act(x, training=training)
            depthwise_output = x

        if self._squeeze_excitation is not None:
            x = self._squeeze_excitation(x)

        x = self._conv3_bn(x, training=training)

        if (self._use_residual and
                self._in_filters == self._out_filters):
            x = tf.add(x, shortcut)

        return x, depthwise_output



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


__ignore__ = r"""


LOCAL_WORK_DPATH=$HOME/temp/debug_smartflow_metrics/ingress
mkdir -p $LOCAL_WORK_DPATH
cd $LOCAL_WORK_DPATH
docker run \
    --runtime=nvidia \
    --volume "$LOCAL_WORK_DPATH":/tmp/ingress \
    --volume $HOME/.aws:/root/.aws:ro \
    --volume "$HOME/code":/extern_code:ro \
    --volume "$HOME/data":/extern_data:ro \
    --volume "$HOME"/.cache/pip:/pip_cache \
    --env AWS_PROFILE=iarpa \
    -it registry.smartgitlab.com/kitware/geowatch:0.16.2-17a6765ff-strict-pyenv3.11.2-20240410T192350-0400-from-0da55667 bash



from geowatch.cli.smartflow.run_iarpa_metrics import *  # NOQA"
config = RunIARPAMetricsCLI(**{
    'region_id': 'KR_R002',
    'input_region_path': 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/split_input/52SDG98/region_models/KR_R002.geojson',
    'true_annot_dpath': 's3://smart-imagery/annotations',
    'pred_site_dpath': 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/consolidated_output_sv/site_models',
    'outbucket': 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/metrics_sv',
    'output_path': 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/metrics_sv/items.jsonl',
    'aws_profile': None,
})

cls = RunIARPAMetricsCLI
cmdline = 0
kwargs = dict(config)
"""


class RunIARPAMetricsCLI(scfg.DataConfig):
    """
    This is an entrypoint to run the IARPA metrics code with Kitware
    modifications, this should be used for debugging and internal metrics, the
    official IARPA image should be used for final scores, and ideally these
    will match results produced here.

    Note:
        This currently runs based on paths and not STAC catalogs, which is an
        issue that would be nice to fix.
    """
    region_id = None
    input_region_path = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))
    true_annot_dpath = None
    pred_site_dpath = None
    outbucket = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))

    output_path = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))

    # input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
    #         '''
    #         Path to input T&E Baseline Framework JSON
    #         '''))
    # input_region_models_asset_name = scfg.Value('cropped_region_models_sc', type=str, required=False, help=ub.paragraph(
    #         '''
    #         Which region model assets to ingress and fix up
    #         '''), alias=['region_models_asset_name'])
    # input_site_models_asset_name = scfg.Value('cropped_site_models_sc', type=str, required=False, help=ub.paragraph(
    #     '''
    #     Which site model assets to ingress and fix up
    #     '''), alias=['site_models_asset_name'])

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> from geowatch.cli.smartflow.run_metrics import *  # NOQA
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> cls = RunIARPAMetricsCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        print('config = ' + ub.urepr(config, nl=1))

        from geowatch.utils.util_framework import NodeStateDebugger
        node_state = NodeStateDebugger()
        node_state.print_environment()
        node_state.print_local_invocation(config)

        from geowatch.utils.util_framework import download_region
        from geowatch.mlops import smart_pipeline
        from geowatch.utils.util_fsspec import FSPath

        # 1. Ingress data
        print("* Running baseline framework kwcoco ingress *")
        ingress_dir = ub.Path('/tmp/ingress')

        eval_dpath = (ingress_dir / 'metrics_output').ensuredir()
        viz_output_dpath = (eval_dpath / 'region_viz_overall').ensuredir()

        USE_NON_STAC_PATHS = True
        if USE_NON_STAC_PATHS:
            # FIXME: would be better if we conformed to the same STAC-in
            # STAC-out paradigm everything else uses, but for now
            # use raw S3 paths.
            remote_true_annot_dpath = FSPath.coerce(config.true_annot_dpath)
            remote_true_region_dpath = remote_true_annot_dpath / 'region_models'
            remote_true_site_dpath = remote_true_annot_dpath / 'site_models'

            remote_pred_site_dpath = FSPath.coerce(config.pred_site_dpath)

            # Select the truth sites to pull down
            remote_true_site_fpaths = [p for p in remote_true_site_dpath.ls() if p.name.startswith(config.region_id)]
            remote_true_region_fpaths = [p for p in remote_true_region_dpath.ls() if p.name.startswith(config.region_id)]

            true_annot_dpath = (ingress_dir / 'truth').ensuredir()
            true_region_dpath = (true_annot_dpath / 'region_models').ensuredir()
            true_site_dpath = (true_annot_dpath / 'site_models').ensuredir()

            # Copy predictions to local node
            pred_annot_dpath = (ingress_dir / 'pred').ensuredir()
            pred_site_dpath = FSPath.coerce(pred_annot_dpath / 'site_models')
            remote_pred_site_dpath.copy(pred_site_dpath, verbose=3)

            # We dont need the region model for scoring, but it helps for
            # visualization and ensuring our approch is consistent
            remote_pred_region_dpath = remote_pred_site_dpath.parent / 'region_models'
            pred_region_dpath = FSPath.coerce(pred_annot_dpath / 'region_models')
            remote_pred_region_dpath.copy(pred_region_dpath)

            # Copy truth to local node
            for fpath in ub.ProgIter(remote_true_site_fpaths, desc='pull site truth', verbose=3):
                fpath.copy(true_site_dpath / fpath.name)
            for fpath in ub.ProgIter(remote_true_region_fpaths, desc='pull region truth', verbose=3):
                fpath.copy(true_region_dpath / fpath.name)

        # 2. Download and prune region file
        print("* Downloading and pruning region file *")
        local_region_path = '/tmp/region.json'
        download_region(
            input_region_path=config.input_region_path,
            output_region_path=local_region_path,
            aws_profile=config.aws_profile,
            strip_nonregions=True,
        )

        node_state.print_current_state(ingress_dir)

        DRAW_SANITY_CHECK = True
        if DRAW_SANITY_CHECK:
            print('Draw predictions and truth before we do eval as a sanity check')
            command = ub.paragraph(
                f'''
                geowatch draw_region {true_site_dpath}
                    --extra_header "True Sites"
                    --fpath "{viz_output_dpath}"/true_site_viz.png
                ''')
            ub.cmd(command, shell=True, verbose=3)

            command = ub.paragraph(
                f'''
                geowatch draw_region {true_region_dpath}
                    --extra_header "True Region"
                    --fpath "{viz_output_dpath}"/true_region_viz.png
                ''')
            ub.cmd(command, shell=True, verbose=3)

            command = ub.paragraph(
                f'''
                geowatch draw_region {pred_site_dpath}
                    --extra_header "Pred Sites"
                    --fpath "{viz_output_dpath}"/pred_site_viz.png
                ''')
            ub.cmd(command, shell=True, verbose=3)

            command = ub.paragraph(
                f'''
                geowatch draw_region {pred_region_dpath}
                    --extra_header "Pred Regions"
                    --fpath "{viz_output_dpath}"/pred_region_viz.png
                ''')
            ub.cmd(command, shell=True, verbose=3)

        smart_pipeline.PolygonEvaluation.name = 'poly_eval'
        eval_node = smart_pipeline.PolygonEvaluation()
        eval_fpath = eval_dpath / 'poly_eval.json'

        eval_node.configure({
            'sites_fpath': pred_site_dpath,  # note this is a bad name, should be "pred_sites"
            'true_site_dpath': true_site_dpath,
            'true_region_dpath': true_region_dpath,
            'eval_dpath': eval_dpath,
            'eval_fpath': eval_fpath,
            'enable_viz': True,  # todo: expose
        })
        command = eval_node.command().rstrip('\\')
        print(command)
        ub.cmd(command, check=True, verbose=3, system=True)

        node_state.print_current_state(ingress_dir)

        node_state.print_directory_contents(eval_dpath)

        # Change viz symlinks to real files
        if True:
            for path in viz_output_dpath.glob('*'):
                if path.is_symlink():
                    real_path = path.resolve()
                    assert not real_path.is_symlink()
                    path.unlink()
                    real_path.copy(path)

        if True:
            confusion_out_dpath = eval_dpath / 'confusion_analysis'
            region_id = config.region_id
            bas_metric_dpath = (eval_dpath / region_id / 'overall/bas')
            src_kwcoco_fpath = None  # TODO: get imagery
            command = ub.codeblock(
                fr'''
                python -m geowatch.mlops.confusor_analysis \
                    --src_kwcoco="{src_kwcoco_fpath}" \
                    --bas_metric_dpath="{bas_metric_dpath}" \
                    --pred_sites="{pred_site_dpath}" \
                    --out_dpath={confusion_out_dpath} \
                    --true_region_dpath="{true_region_dpath}" \
                    --true_site_dpath="{true_site_dpath}" \
                    --region_id="{region_id}" \
                    --viz_sites=0 \
                    --reload=0
                ''')
            ub.cmd(command, verbose=3, shell=1)

            confusion_group_dpath = confusion_out_dpath / 'confusion_groups'
            confusion_region_fpaths = list(confusion_group_dpath.glob('*.geojson'))
            for fpath in confusion_region_fpaths:
                command = ub.paragraph(
                    f'''
                    geowatch draw_region {fpath}
                        --extra_header "{fpath.name}"
                        --fpath "{fpath}.png"
                    ''')
                ub.cmd(command, shell=True, verbose=3)

        assets_to_egress = {
            'eval_dpath': eval_dpath,
            'eval_fpath': eval_fpath,
        }
        from geowatch.cli.smartflow_egress import smartflow_egress
        smartflow_egress(assets_to_egress,
                         local_region_path,
                         config.output_path,
                         config.outbucket,
                         aws_profile=config.aws_profile,
                         # dryrun=config.dryrun,
                         # newline=config.newline
                         )


__cli__ = RunIARPAMetricsCLI
main = __cli__.main

if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/geowatch/geowatch/cli/smartflow/run_metrics.py
        python -m geowatch.cli.smartflow.run_metrics
    """
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains code to build an ASPP layer.

Reference:
  - [Rethinking Atrous Convolution for Semantic Image Segmentation](
      https://arxiv.org/pdf/1706.05587.pdf)
  - [ParseNet: Looking Wider to See Better](
      https://arxiv.org/pdf/1506.04579.pdf).
"""
import tensorflow as tf
from absl import logging
from deeplab2.model import utils
from deeplab2.model.layers import convolutions

layers = tf.keras.layers
backend = tf.keras.backend


class ASPPConv(tf.keras.layers.Layer):
    """An atrous convolution for ASPP."""

    def __init__(self,
                 output_channels,
                 atrous_rate,
                 name,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu'):
        """Creates a atrous convolution layer for the ASPP.

        This layer consists of an atrous convolution followed by a BatchNorm layer
        and a ReLU activation.

        Args:
          output_channels: An integer specifying the number of output channels of
            the convolution.
          atrous_rate: An integer specifying the atrous/dilation rate of the
            convolution.
          name: A string specifying the name of this layer.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
        """
        super(ASPPConv, self).__init__(name=name)

        self._conv_bn_act = convolutions.Conv2DSame(
            output_channels,
            kernel_size=3,
            name='conv_bn_act',
            atrous_rate=atrous_rate,
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=activation)

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        return self._conv_bn_act(input_tensor, training=training)


class ASPPPool(tf.keras.layers.Layer):
    """A pooling layer for ASPP."""

    def __init__(self,
                 output_channels,
                 name,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu'):
        """Creates a pooling layer for the ASPP.

        This layer consists of a global average pooling, followed by a convolution,
        and by a BatchNorm layer and a ReLU activation.

        Args:
          output_channels: An integer specifying the number of output channels of
            the convolution.
          name: A string specifying the name of this layer.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
        """
        super(ASPPPool, self).__init__(name=name)

        self._pool_size = (None, None)
        self._conv_bn_act = convolutions.Conv2DSame(
            output_channels,
            kernel_size=1,
            name='conv_bn_act',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=activation)

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the pooling layer.

        The default behavior of the pooling layer is global average pooling. A
        custom pooling size can be set here.

        Args:
          pool_size: A tuple specifying the pooling size of the pooling layer.

        Raises:
          An error occurs if exactly one pooling dimension is set to 'None'.
        """
        # If exactly one pooling dimension is 'None' raise an error.
        if None in pool_size and pool_size != (None, None):
            raise ValueError('The ASPP pooling layer requires that the pooling size '
                             'is set explicitly for both dimensions. In case, global '
                             'average pooling should be used, call '
                             'reset_pooling_layer() or set both to None.')

        self._pool_size = pool_size
        logging.info('Global average pooling in the ASPP pooling layer was replaced'
                     ' with tiled average pooling using the provided pool_size. '
                     'Please make sure this behavior is intended.')

    def get_pool_size(self):
        return self._pool_size

    def reset_pooling_layer(self):
        """Resets the pooling layer to global average pooling."""
        self._pool_size = (None, None)

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        if tuple(self._pool_size) == (None, None):
            # Global image pooling
            pool_size = input_tensor.shape[1:3]
        else:
            # Tiled image pooling
            pool_size = self._pool_size

        x = backend.pool2d(input_tensor, pool_size, padding='valid',
                           pool_mode='avg')
        x = self._conv_bn_act(x, training=training)

        target_h = tf.shape(input_tensor)[1]
        target_w = tf.shape(input_tensor)[2]

        x = utils.resize_align_corners(x, [target_h, target_w])
        return x


class ASPP(tf.keras.layers.Layer):
    """An atrous spatial pyramid pooling layer."""

    def __init__(self,
                 output_channels,
                 atrous_rates,
                 aspp_use_only_1x1_proj_conv=False,
                 name='ASPP',
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu'):
        """Creates an ASPP layer.

        Args:
          output_channels: An integer specifying the number of output channels of
            each ASPP convolution layer.
          atrous_rates: A list of three integers specifying the atrous/dilation rate
            of each ASPP convolution layer.
          aspp_use_only_1x1_proj_conv: Boolean, specifying if the ASPP five branches
            are turned off or not. If True, the ASPP module is degenerated to one
            1x1 convolution, projecting the input channels to `output_channels`.
          name: A string specifying the name of this layer (default: 'ASPP').
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.

        Raises:
          ValueError: An error occurs when atrous_rates does not contain 3
            elements and `aspp_use_only_1x1_proj_conv` is False.
        """
        super(ASPP, self).__init__(name=name)

        if not aspp_use_only_1x1_proj_conv and len(atrous_rates) != 3:
            raise ValueError(
                'The ASPP layers need exactly 3 atrous rates, but %d were given' %
                len(atrous_rates))
        self._aspp_use_only_1x1_proj_conv = aspp_use_only_1x1_proj_conv

        # Projection convolution is always used.
        self._proj_conv_bn_act = convolutions.Conv2DSame(
            output_channels,
            kernel_size=1,
            name='proj_conv_bn_act',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=activation)

        if not aspp_use_only_1x1_proj_conv:
            self._conv_bn_act = convolutions.Conv2DSame(
                output_channels,
                kernel_size=1,
                name='conv_bn_act',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation=activation)
            rate1, rate2, rate3 = atrous_rates
            self._aspp_conv1 = ASPPConv(output_channels, rate1, name='aspp_conv1',
                                        bn_layer=bn_layer, activation=activation)
            self._aspp_conv2 = ASPPConv(output_channels, rate2, name='aspp_conv2',
                                        bn_layer=bn_layer, activation=activation)
            self._aspp_conv3 = ASPPConv(output_channels, rate3, name='aspp_conv3',
                                        bn_layer=bn_layer, activation=activation)
            self._aspp_pool = ASPPPool(output_channels, name='aspp_pool',
                                       bn_layer=bn_layer, activation=activation)
            # Dropout is needed only when ASPP five branches are used.
            self._proj_drop = layers.Dropout(rate=0.1)

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layer.

        The default behavior of the pooling layer is global average pooling. A
        custom pooling size can be set here.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layer.
        """
        if not self._aspp_use_only_1x1_proj_conv:
            self._aspp_pool.set_pool_size(pool_size)

    def get_pool_size(self):
        if not self._aspp_use_only_1x1_proj_conv:
            return self._aspp_pool.get_pool_size()
        else:
            return (None, None)

    def reset_pooling_layer(self):
        """Resets the pooling layer to global average pooling."""
        self._aspp_pool.reset_pooling_layer()

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        if self._aspp_use_only_1x1_proj_conv:
            x = self._proj_conv_bn_act(input_tensor, training=training)
        else:
            # Apply the ASPP module.
            results = []
            results.append(self._conv_bn_act(input_tensor, training=training))
            results.append(self._aspp_conv1(input_tensor, training=training))
            results.append(self._aspp_conv2(input_tensor, training=training))
            results.append(self._aspp_conv3(input_tensor, training=training))
            results.append(self._aspp_pool(input_tensor, training=training))
            x = tf.concat(results, 3)
            x = self._proj_conv_bn_act(x, training=training)
            x = self._proj_drop(x, training=training)
        return x



import ubelt as ub

try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


class TrackFunction:
    """
    Abstract class that all track functions should inherit from.
    """

    def __call__(self, sub_dset, video_id, **kwargs):
        # The original impl with __call__ methods made it difficult for me to
        # grep for things, so I'm disabling it.
        raise AssertionError("Use the explicit .forward method instead")
        return self.forward(sub_dset, video_id,  **kwargs)

    def forward(self, sub_dset, video_id):
        """
        Detect new annotations and track them in images for a single video.

        Args:
            sub_dset (CocoDataset):
            video_id (int): video to track

        Returns:
            kwcoco.CocoDataset
        """
        raise NotImplementedError('must be implemented by subclasses')

    def apply_per_video(self, coco_dset):
        """
        Main entrypoint for this class.

        Calls :func:`safe_apply` on each video in the coco dataset.

        Args:
            coco_dset (kwcoco.CocoDataset):
                the dataset to run tracking on

        Returns:
            kwcoco.CocoDataset:
                A modified or copied version of the input with new annotations
                and tracks.

        Example:
            >>> from geowatch.tasks.tracking.abstract_classes import *  # NOQA
            >>> import kwcoco
            >>> self = NoOpTrackFunction()
            >>> coco_dset = kwcoco.CocoDataset.coerce('vidshapes8')
            >>> self.apply_per_video(coco_dset)
        """
        import kwcoco
        from geowatch.utils import kwcoco_extensions
        import rich

        # Note (2024-01-24): previously this was implemented in a way that
        # broke up a larger coco dataset into one for each video, but we have
        # modified the implementation to avoid the "subset" and "union"
        # overhead.

        # If there are downstream issues set to True for old behavior
        # otherwise, when stable the else branch should be deleted.
        DO_SUBSET_HACK = False  # True for old behavior

        tracking_results = []
        vid_gids = list(coco_dset.index.vidid_to_gids.items())
        total = len(coco_dset.index.vidid_to_gids)
        for video_id, gids in ub.ProgIter(vid_gids,
                                          total=total,
                                          desc='apply_per_video',
                                          verbose=3):

            rich.print(f'[blue]-- Apply Tracking Function to Video {video_id}')

            # Beware, in the past there was a crash here that required
            # wrapping the rest of this loop in a try/except. -csg
            sub_dset = self.safe_apply(coco_dset, video_id, gids,
                                       DO_SUBSET_HACK=DO_SUBSET_HACK)

            # Store a reference to the dataset (which may be a modified subset,
            # or an unmodified reference with the image ids that were tracked)
            tracking_results.append({
                'sub_dset': sub_dset,
                'video_id': video_id,
                'image_ids': gids,
            })
            rich.print('[blue]----')

        # Tracks were either updated or added.
        # In the case they were updated the existing track ids should
        # be disjoint. All new tracks should not overlap with

        _debug = 0

        new_trackids = kwcoco_extensions.TrackidGenerator(None)
        for tracking_result in ub.ProgIter(tracking_results,
                                           desc='Ensure ok tracks',
                                           verbose=3):

            sub_dset = tracking_result['sub_dset']
            sub_gids = tracking_result['image_ids']

            # Rebuild the index to ensure any hacks are removed.
            # We should be able to remove this step.
            # sub_dset._build_index()

            sub_aids = list(ub.flatten(sub_dset.images(sub_gids).annots))
            sub_annots = sub_dset.annots(sub_aids)
            sub_tids = sub_annots.lookup('track_id')
            existing_tids = set(sub_tids)

            collisions = existing_tids & new_trackids.used_trackids
            if _debug:
                print('existing_tids = {!r}'.format(existing_tids))
                print('collisions = {!r}'.format(collisions))

            new_trackids.exclude_trackids(existing_tids)
            if collisions:
                old_tid_to_aids = ub.group_items(sub_annots, sub_tids)
                assert len(old_tid_to_aids) == len(existing_tids)
                print(f'Resolve {len(collisions)} track-id collisions')
                # Change the track ids of any collisions
                for old_tid in collisions:
                    new_tid = next(new_trackids)
                    # Note: this does not update the index, but we
                    # are about to clobber it anyway, so it doesnt matter
                    for aid in old_tid_to_aids[old_tid]:
                        ann = sub_dset.index.anns[aid]
                        ann['track_id'] = new_tid
                    existing_tids.add(new_tid)
            new_trackids.exclude_trackids(existing_tids)

            if _debug:
                after_tids = set(sub_annots.lookup('track_id'))
                print('collisions = {!r}'.format(collisions))
                print(f'{after_tids=}')

        if DO_SUBSET_HACK:
            # If we broke up the dataset into subsets, we need to combine them
            # back together.
            fixed_subdataset = [r['sub_dset'] for r in tracking_results]
            coco_dset = kwcoco.CocoDataset.union(*fixed_subdataset,
                                                 disjoint_tracks=False)

        if _debug:
            x = coco_dset.annots().images.get('video_id')
            y = coco_dset.annots().get('track_id')
            z = ub.group_items(x, y)
            track_to_num_videos = ub.map_vals(set, z)
            if track_to_num_videos:
                assert max(map(len, track_to_num_videos.values())) == 1, (
                    'track belongs to multiple videos!')
        return coco_dset

    @profile
    def safe_apply(self, coco_dset, video_id, gids, DO_SUBSET_HACK=False):
        import numpy as np
        DEBUG_JSON_SERIALIZABLE = 0
        if DEBUG_JSON_SERIALIZABLE:
            from kwutil.util_json import debug_json_unserializable

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(coco_dset.dataset,
                                      'Input to safe_apply: ')

        if DO_SUBSET_HACK:
            # Try not to do this if we can avoid it.
            sub_dset = self.safe_partition(coco_dset, gids)
        else:
            # Simulate a single-video dataset (and maintain relevant caches)
            sub_dset = coco_dset
            # TODO: holding context to revent re-looking up annotations for
            # each video-id will likely improve performance.
            # sub_dset = WrappedSingleVideoCocoDataset(coco_dset, video_id, gids)

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(sub_dset.dataset, 'Before __call__')

        orig_aids = list(ub.flatten(sub_dset.images(gids).annots))
        orig_annots = sub_dset.annots(orig_aids)
        # orig_annots = sub_dset.annots()
        orig_tids = orig_annots.get('track_id', None)
        orig_trackless_flags = np.array([tid is None for tid in orig_tids])
        orig_aids = list(orig_annots)

        ####
        # APPLY THE TRACKING FUNCTION.
        # THIS IS THE MAIN WORK. SEE SPECIFIC forward FUNCTIONS
        sub_dset = self.forward(sub_dset, video_id)
        ####

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(sub_dset.dataset, 'After __call__')

        # if new annots were not created, rollover the old tracks
        new_aids = list(ub.flatten(sub_dset.images(gids).annots))
        new_annots = sub_dset.annots(new_aids)
        if new_annots.aids == orig_aids:
            new_tids = new_annots.get('track_id', None)
            # Only overwrite track ids for annots that didn't have them
            new_tids = np.where(orig_trackless_flags, new_tids, orig_tids)

            # Ensure types are json serializable
            import numbers

            def _fixtype(tid):
                # need to keep strings the same, but integers need to be
                # cast from numpy to python ints.
                if isinstance(tid, numbers.Integral):
                    return int(tid)
                else:
                    return tid

            new_tids = list(map(_fixtype, new_tids))

            new_annots.set('track_id', new_tids)

        # TODO: why is this assert here?
        sub_track_ids = new_annots.lookup('track_id', None)
        if None in sub_track_ids:
            raise AssertionError(f'None in track ids: {sub_track_ids}')

        out_dset = sub_dset

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(out_dset.dataset,
                                      'Output of safe_apply: ')
        return out_dset

    @staticmethod
    @profile
    def safe_partition(coco_dset, gids):

        sub_dset = coco_dset.subset(gids=gids, copy=True)
        # HACK ensure tracks are not duplicated between videos
        # (if they are, this is fixed in dedupe_tracks anyway)
        sub_dset.index.trackid_to_aids.update(coco_dset.index.trackid_to_aids)
        return sub_dset

    @staticmethod
    @profile
    def safe_union(coco_dset, new_dset, existing_aids=[]):
        raise AssertionError('scheduled for removal')
        coco_dset._build_index()
        new_dset._build_index()
        # we handle tracks in normalize.dedupe_tracks anyway, and
        # disjoint_tracks=True interferes with keeping site_ids around as
        # track_ids.
        # return coco_dset.union(new_dset, disjoint_tracks=True)
        return coco_dset.union(new_dset, disjoint_tracks=False)


class NoOpTrackFunction(TrackFunction):
    """
    Use existing tracks.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs  # Unused

    def forward(self, sub_dset, video_id):
        return sub_dset


class NewTrackFunction(TrackFunction):
    """
    Specialization of TrackFunction to create polygons that do not yet exist
    in coco_dset, and add them as new annotations
    """

    def forward(self, sub_dset, video_id):
        # print(f'Enter {self.__class__} __call__ function')
        # print('Create tracks')
        tracks = self.create_tracks(sub_dset, video_id)
        # print('Add tracks to dset')
        sub_dset = self.add_tracks_to_dset(sub_dset, tracks)
        # print('After tracking sub_dset.stats(): ' +
        #       ub.urepr(sub_dset.basic_stats()))
        # print(f'Exit {self.__class__} __call__ function')
        return sub_dset

    def create_tracks(self, sub_dset, video_id):
        """
        Args:
            sub_dset (CocoDataset):
            video_id (int): video to create tracks for

        Returns:
            GeoDataFrame
        """
        raise NotImplementedError('must be implemented by subclasses')

    def add_tracks_to_dset(self, sub_dset, tracks):
        """
        Args:
            tracks (GeoDataFrame):

        Returns:
            kwcoco.CocoDataset
        """
        raise NotImplementedError('must be implemented by subclasses')


class WrappedSingleVideoCocoDataset:
    def __init__(self, dset, video_id, gids):
        self.dset = dset
        self.video_id = video_id
        self.gids = gids
